from jesse.strategies import Strategy


class TestPositionWithLeverage2(Strategy):
    def should_long(self) -> bool:
        return self.price == 10

    def should_short(self) -> bool:
        return False

    def go_long(self):
        # price is at $1
        self.buy = 10, self.price

    def update_position(self):
        if self.price == 12:
            assert self.position.exchange.futures_leverage == 2
            assert self.position.pnl == 20
            assert self.position.total_cost == 50
            assert self.position.roi == 40
            assert self.position.pnl_percentage == self.position.roi

    def go_short(self):
        pass

    def should_cancel(self):
        return False



import arrow
import jesse.helpers as jh


class AppState:
    time = arrow.utcnow().int_timestamp * 1000
    starting_time = None
    daily_balance = []

    # used as placeholders for detecting open trades metrics
    total_open_trades = 0
    total_open_pl = 0
    total_liquidations = 0

    session_id = ''

    def set_session_id(self) -> None:
        """
        Generated and sets session_id. Used to prevent overriding of the session_id
        """
        if self.session_id == '':
            self.session_id = jh.generate_unique_id()



class sides:
    BUY = 'buy'
    SELL = 'sell'


class trade_types:
    LONG = 'long'
    SHORT = 'short'


class order_statuses:
    ACTIVE = 'ACTIVE'
    CANCELED = 'CANCELED'
    EXECUTED = 'EXECUTED'
    PARTIALLY_EXECUTED = 'PARTIALLY EXECUTED'
    QUEUED = 'QUEUED'
    LIQUIDATED = 'LIQUIDATED'


class timeframes:
    MINUTE_1 = '1m'
    MINUTE_3 = '3m'
    MINUTE_5 = '5m'
    MINUTE_15 = '15m'
    MINUTE_30 = '30m'
    MINUTE_45 = '45m'
    HOUR_1 = '1h'
    HOUR_2 = '2h'
    HOUR_3 = '3h'
    HOUR_4 = '4h'
    HOUR_6 = '6h'
    HOUR_8 = '8h'
    HOUR_12 = '12h'
    DAY_1 = '1D'


class colors:
    GREEN = 'green'
    YELLOW = 'yellow'
    RED = 'red'
    MAGENTA = 'magenta'
    BLACK = 'black'


class order_roles:
    OPEN_POSITION = 'OPEN POSITION'
    CLOSE_POSITION = 'CLOSE POSITION'
    INCREASE_POSITION = 'INCREASE POSITION'
    REDUCE_POSITION = 'REDUCE POSITION'


class order_flags:
    OCO = 'OCO'
    POST_ONLY = 'PostOnly'
    CLOSE = 'Close'
    HIDDEN = 'Hidden'
    REDUCE_ONLY = 'ReduceOnly'


class order_types:
    MARKET = 'MARKET'
    LIMIT = 'LIMIT'
    STOP = 'STOP'
    FOK = 'FOK'
    STOP_LIMIT = 'STOP LIMIT'


class exchanges:
    COINBASE = 'Coinbase'
    BITFINEX = 'Bitfinex'
    BINANCE = 'Binance'
    BINANCE_FUTURES = 'Binance Futures'
    TESTNET_BINANCE_FUTURES = 'Testnet Binance Futures'
    SANDBOX = 'Sandbox'



import jesse.helpers as jh
from jesse.services import logger as jesse_logger
import threading
import traceback
from jesse.services.redis import sync_publish


def register_custom_exception_handler() -> None:
    # other threads
    def handle_thread_exception(args) -> None:
        if args.exc_type == SystemExit:
            return

        if args.exc_type.__name__ == 'Termination':
            sync_publish('termination', {})
            jh.terminate_app()
        else:
            # send notifications if it's a live session
            if jh.is_live():
                jesse_logger.error(
                    f'{args.exc_type.__name__}: {args.exc_value}'
                )

            sync_publish('exception', {
                'error': f"{args.exc_type.__name__}: {str(args.exc_value)}",
                'traceback': str(traceback.format_exc())
            })
            terminate_session()

    threading.excepthook = handle_thread_exception


def terminate_session():
    sync_publish('unexpectedTermination', {
        'message': "Session terminated as the result of an uncaught exception",
    })

    jesse_logger.error('Session terminated as the result of an uncaught exception')

    jh.terminate_app()



import jesse.helpers as jh
from jesse.models.utils import store_daily_balance_into_db
from jesse.services import logger
from jesse.store import store


def save_daily_portfolio_balance() -> None:
    # # store daily_balance of assets into database
    # if jh.is_livetrading():
    #     for asset_key, asset_value in e.assets.items():
    #         store_daily_balance_into_db({
    #             'id': jh.generate_unique_id(),
    #             'timestamp': jh.now(),
    #             'identifier': jh.get_config('env.identifier', 'main'),
    #             'exchange': e.name,
    #             'asset': asset_key,
    #             'balance': asset_value,
    #         })
    balances = []
    for key, e in store.exchanges.storage.items():
        try:
            balances.append(
                e.assets[jh.app_currency()]
            )
        except KeyError:
            raise ValueError('Invalid quote trading pair. Check your trading route\'s symbol')

    # add open position values
    for key, pos in store.positions.storage.items():
        if pos.is_open:
            balances.append(pos.pnl)

    total = sum(balances)
    store.app.daily_balance.append(total)

    # TEMP: disable storing in database for now
    if not jh.is_livetrading():
        logger.info(f'Saved daily portfolio balance: {round(total, 2)}')



import math
from decimal import Decimal
from typing import Union

import numpy as np
import pandas as pd

import jesse.helpers as jh
from jesse.enums import timeframes


def anchor_timeframe(timeframe: str) -> str:
    """
    Returns the anchor timeframe. Useful for writing
    dynamic strategies using multiple timeframes.

    :param timeframe: str
    :return: str
    """

    dic = {
        timeframes.MINUTE_1: timeframes.MINUTE_5,
        timeframes.MINUTE_3: timeframes.MINUTE_15,
        timeframes.MINUTE_5: timeframes.MINUTE_30,
        timeframes.MINUTE_15: timeframes.HOUR_2,
        timeframes.MINUTE_30: timeframes.HOUR_3,
        timeframes.MINUTE_45: timeframes.HOUR_3,
        timeframes.HOUR_1: timeframes.HOUR_4,
        timeframes.HOUR_2: timeframes.HOUR_6,
        timeframes.HOUR_3: timeframes.DAY_1,
        timeframes.HOUR_4: timeframes.DAY_1,
        timeframes.HOUR_6: timeframes.DAY_1,
        timeframes.HOUR_8: timeframes.DAY_1,
        timeframes.HOUR_12: timeframes.DAY_1,
    }

    return dic[timeframe]


def crossed(series1: np.ndarray, series2: Union[float, int, np.ndarray], direction: str = None,
            sequential: bool = False) -> bool:
    """
    Helper for detection of crosses

    :param series1: np.ndarray
    :param series2: float, int, np.array
    :param direction: str - default: None - above or below
    :param sequential: bool - default: False

    :return: bool
    """

    if sequential:
        series1_shifted = jh.np_shift(series1, 1, np.nan)

        if type(series2) is np.ndarray:
            series2_shifted = jh.np_shift(series2, 1, np.nan)
        else:
            series2_shifted = series2

        if direction is None or direction == "above":
            cross_above = np.logical_and(series1 > series2, series1_shifted <= series2_shifted)

        if direction is None or direction == "below":
            cross_below = np.logical_and(series1 < series2, series1_shifted >= series2_shifted)

        if direction is None:
            return np.logical_or(cross_above, cross_below)

    else:
        if type(series2) is not np.ndarray:
            series2 = np.array([series2, series2])

        if direction is None or direction == "above":
            cross_above = series1[-2] <= series2[-2] and series1[-1] > series2[-1]
        if direction is None or direction == "below":
            cross_below = series1[-2] >= series2[-2] and series1[-1] < series2[-1]

        if direction is None:
            return cross_above or cross_below

    if direction == "above":
        return cross_above
    else:
        return cross_below


def estimate_risk(entry_price: float, stop_price: float) -> float:
    """
    estimates the risk per share

    :param entry_price: float
    :param stop_price: float
    :return: float
    """
    if math.isnan(entry_price) or math.isnan(stop_price):
        raise TypeError()

    return abs(entry_price - stop_price)


def limit_stop_loss(entry_price: float, stop_price: float, trade_type: str, max_allowed_risk_percentage: int) -> float:
    """
    Limits the stop-loss price according to the max allowed risk percentage.
    (How many percent you're OK with the price going against your position)

    :param entry_price:
    :param stop_price:
    :param trade_type:
    :param max_allowed_risk_percentage:
    :return: float
    """
    risk = abs(entry_price - stop_price)
    max_allowed_risk = entry_price * (max_allowed_risk_percentage / 100)
    risk = min(risk, max_allowed_risk)
    return (entry_price - risk) if trade_type == 'long' else (entry_price + risk)


def numpy_candles_to_dataframe(candles: np.ndarray, name_date: str = "date", name_open: str = "open",
                               name_high: str = "high",
                               name_low: str = "low", name_close: str = "close",
                               name_volume: str = "volume") -> pd.DataFrame:
    columns = [name_date, name_open, name_close, name_high, name_low, name_volume]
    df = pd.DataFrame(data=candles, index=pd.to_datetime(candles[:, 0], unit="ms"), columns=columns)
    df[name_date] = pd.to_datetime(df.index, unit="ms")
    return df


def qty_to_size(qty: float, price: float) -> float:
    """
    converts quantity to position-size
    example: requesting 2 shares at the price of %50 would return $100

    :param qty: float
    :param price: float
    :return: float
    """
    if math.isnan(qty) or math.isnan(price):
        raise TypeError()

    return qty * price


def risk_to_qty(capital: float, risk_per_capital: float, entry_price: float, stop_loss_price: float, precision: int = 8,
                fee_rate: float = 0) -> float:
    """
    a risk management tool to quickly get the qty based on risk percentage

    :param capital:
    :param risk_per_capital:
    :param entry_price:
    :param stop_loss_price:
    :param precision:
    :param fee_rate:
    :return: float
    """
    risk_per_qty = abs(entry_price - stop_loss_price)
    size = risk_to_size(capital, risk_per_capital, risk_per_qty, entry_price)

    if fee_rate != 0:
        size = size * (1 - fee_rate * 3)

    return size_to_qty(size, entry_price, precision=precision, fee_rate=fee_rate)


def risk_to_size(capital_size: float, risk_percentage: float, risk_per_qty: float, entry_price: float) -> float:
    """
    calculates the size of the position based on the amount of risk percentage you're willing to take
    example: round(risk_to_size(10000, 1, 0.7, 8.6)) == 1229

    :param capital_size:
    :param risk_percentage:
    :param risk_per_qty:
    :param entry_price:
    :return: float
    """
    if risk_per_qty == 0:
        raise ValueError('risk cannot be zero')

    risk_percentage /= 100
    temp_size = ((risk_percentage * capital_size) / risk_per_qty) * entry_price
    return min(temp_size, capital_size)


def size_to_qty(position_size: float, entry_price: float, precision: int = 3, fee_rate: float = 0) -> float:
    """
    converts position-size to quantity
    example: requesting $100 at the entry_price of $50 would return 2
    :param position_size: float
    :param entry_price: float
    :param precision: int
    :param fee_rate:
    :return: float
    """
    if math.isnan(position_size) or math.isnan(entry_price):
        raise TypeError()

    if fee_rate != 0:
        position_size *= 1 - fee_rate * 3

    return jh.floor_with_precision(position_size / entry_price, precision)


def subtract_floats(float1: float, float2: float) -> float:
    """
    Subtracts two floats without the rounding issue in Python

    :param float1: float
    :param float2: float

    :return: float
    """
    return float(Decimal(str(float1)) - Decimal(str(float2)))


def sum_floats(float1: float, float2: float) -> float:
    """
    Sums two floats without the rounding issue in Python

    :param float1: float
    :param float2: float

    :return: float
    """
    return float(Decimal(str(float1)) + Decimal(str(float2)))


def strictly_increasing(series: np.ndarray, lookback: int) -> bool:
    a = series[-lookback:]
    diff = np.diff(a)
    return np.all(diff > 0)


def strictly_decreasing(series: np.ndarray, lookback: int) -> bool:
    a = series[-lookback:]
    diff = np.diff(a)
    return np.all(diff < 0)


def streaks(series: np.ndarray, use_diff=True) -> np.ndarray:
    if use_diff:
        series = np.diff(series)
    pos = np.clip(series, 0, 1).astype(bool).cumsum()
    neg = np.clip(series, -1, 0).astype(bool).cumsum()
    streak = np.where(series >= 0, pos - np.maximum.accumulate(np.where(series <= 0, pos, 0)),
                      -neg + np.maximum.accumulate(np.where(series >= 0, neg, 0)))

    return np.concatenate(
        (np.full((series.shape[0] - streak.shape[0]), np.nan), streak)
    )


def signal_line(series: np.ndarray, period: int = 10, matype: int = 0) -> np.ndarray:
    from jesse.indicators.ma import ma
    return ma(series, period=period, matype=matype, sequential=True)


def kelly_criterion(win_rate: float, ratio_avg_win_loss: float) -> float:
    return win_rate - ((1 - win_rate) / ratio_avg_win_loss)


def prices_to_returns(price_series: np.ndarray) -> np.ndarray:
    """
    converts a series of asset prices to returns.
    """
    pct = np.diff(price_series) / price_series[:-1] * 100
    return jh.same_length(price_series, pct)


def z_score(series: np.ndarray) -> np.ndarray:
    """
    A Z-score is a numerical measurement that describes a value's relationship to the mean of a group of values. Z-score is measured in terms of standard deviations from the mean
    """
    return (series - np.mean(series)) / np.std(series)


def are_cointegrated(price_returns_1: np.ndarray, price_returns_2: np.ndarray, cutoff=0.05) -> bool:
    """
    Uses unit-root test on residuals to test for cointegrated relationship
    See Hamilton (1994) 19.2 for more details
    H_0 is that there is no cointegration i.e. that the residuals have are unit root series (non-stationary)
    We must observe significant p-value to convince ourselves that the series are cointegrated
    """
    from statsmodels.tsa.stattools import coint

    p_value = coint(price_returns_1, price_returns_2)[1]
    return p_value < cutoff


def dd(msg: str) -> None:
    """
    The dd function dumps the given variables and ends execution of the script. 
    Used for debugging. 

    :param msg: str
    """
    print(msg)
    jh.terminate_app()


def combinations_without_repeat(a: np.ndarray, n: int = 2) -> np.ndarray:
    """
    Creates an array containing all combinations of the passed arrays individual values without repetitions. Useful for the optimization mode.
    """
    from itertools import permutations
    if n <= 1:
        raise ValueError("n must be >= 2")
    return np.array(list(permutations(a, n)))


def wavelet_denoising(raw: np.ndarray, wavelet='haar', level: int = 1, mode: str = 'symmetric',
                      smoothing_factor: float = 0, threshold_mode: str = 'hard') -> np.ndarray:
    """
    deconstructs, thresholds then reconstructs
    higher thresholds = less detailed reconstruction

    Only consider haar, db, sym, coif wavelet basis functions, as these are relatively suitable for financial data
    """
    import pywt
    # Deconstruct
    coeff = pywt.wavedec(raw, wavelet, mode=mode)
    # Mean absolute deviation of a signal
    max_level = pywt.dwt_max_level(len(raw), wavelet)
    level = min(level, max_level)
    madev = np.mean(np.absolute(coeff[-level] - np.mean(coeff[-level])))
    # The hardcored factor is explained here: https://en.wikipedia.org/wiki/Median_absolute_deviation
    sigma = (1 / 0.67449) * madev * smoothing_factor
    threshold = sigma * np.sqrt(2 * np.log(len(raw)))
    coeff[1:] = (pywt.threshold(i, value=threshold, mode=threshold_mode) for i in coeff[1:])
    signal = pywt.waverec(coeff, wavelet, mode=mode)
    if len(signal) > len(raw):
        signal = np.delete(signal, -1)
    return signal



from jesse.strategies import Strategy


# test_on_route_close_position part 1 - BTC-USD
class Test25(Strategy):
    def should_long(self):
        # buy on market at first candle, close when on_route_stop_loss event is fired
        return self.index == 0

    def should_short(self):
        return False

    def go_long(self):
        qty = 1
        self.buy = qty, self.price

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def on_route_close_position(self, strategy):
        self.take_profit = 1, self.price



from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source
from jesse.helpers import slice_candles

SINEWAVE = namedtuple('SINEWAVE', ['sine', 'lead'])


def ht_sine(candles: np.ndarray, source_type: str = "close", sequential: bool = False) -> SINEWAVE:
    """
    HT_SINE - Hilbert Transform - SineWave

    :param candles: np.ndarray
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: SINEWAVE(sine, lead)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    sine, leadsine = talib.HT_SINE(source)

    if sequential:
        return SINEWAVE(sine, leadsine)
    else:
        return SINEWAVE(sine[-1], leadsine[-1])



import os

import arrow
import numpy as np
import pytest

import jesse.helpers as jh


def test_app_currency():
    from jesse.routes import router
    from jesse.enums import exchanges, timeframes
    router.initiate(
        [{'exchange': exchanges.BITFINEX, 'symbol': 'ETH-USD', 'timeframe': timeframes.HOUR_3, 'strategy': 'Test19'}])
    assert jh.app_currency() == 'USD'


def test_app_mode():
    assert jh.app_mode() == 'backtest'


def test_arrow_to_timestamp():
    arrow_time = arrow.get('2015-08-01')
    assert jh.arrow_to_timestamp(arrow_time) == 1438387200000


def test_base_asset():
    assert jh.base_asset('BTC-USDT') == 'BTC'
    assert jh.base_asset('BTC-USD') == 'BTC'
    assert jh.base_asset('DEFI-USDT') == 'DEFI'
    assert jh.base_asset('DEFI-USD') == 'DEFI'


def test_binary_search():
    arr = [0, 11, 22, 33, 44, 54, 55]

    assert jh.binary_search(arr, 22) == 2
    assert jh.binary_search(arr, 222) == -1


def test_clean_orderbook_list():
    assert np.array_equal(
        jh.clean_orderbook_list([
            ['10', '11'],
            ['11', '14'],
            ['12', '13'],
            ['13', '133'],
            ['14', '12'],
        ]),
        [
            [10, 11],
            [11, 14],
            [12, 13],
            [13, 133],
            [14, 12],
        ]
    )


def test_color():
    msg_text = 'msg'
    msg_color = 'black'
    assert jh.color(msg_text, msg_color) == '\x1b[30mmsg\x1b[0m'


def test_convert_number():
    old_max = 119
    old_min = 40
    new_max = 4.0
    new_min = 0.5
    old_value = 41

    assert jh.convert_number(old_max, old_min, new_max, new_min, old_value) == 0.5443037974683544


def test_dashless_symbol():
    assert jh.dashless_symbol('BTC-USD') == 'BTCUSD'
    assert jh.dashless_symbol('BTC-USDT') == 'BTCUSDT'
    assert jh.dashless_symbol('1INCH-USDT') == '1INCHUSDT'
    assert jh.dashless_symbol('SC-USDT') == 'SCUSDT'

    # make sure that it works even if it's already dashless
    assert jh.dashless_symbol('BTCUSDT') == 'BTCUSDT'


def test_dashy_symbol():
    assert jh.dashy_symbol('BTCUSD') == 'BTC-USD'
    assert jh.dashy_symbol('BTCUSDT') == 'BTC-USDT'
    assert jh.dashy_symbol('BTC-USDT') == 'BTC-USDT'


def test_date_diff_in_days():
    date_1 = arrow.get('2015-12-23 18:40:48', 'YYYY-MM-DD HH:mm:ss')
    date_2 = arrow.get('2017-11-15 13:18:20', 'YYYY-MM-DD HH:mm:ss')
    diff = jh.date_diff_in_days(date_1, date_2)
    assert diff == 692


def test_date_to_timestamp():
    assert jh.date_to_timestamp('2015-08-01') == 1438387200000


def test_dna_to_hp():
    strategy_hp = [
        {'name': 'hp1', 'type': float, 'min': 0.01, 'max': 1.0, 'default': 0.09},
        {'name': 'hp2', 'type': int, 'min': 1, 'max': 10, 'default': 2},
    ]
    dna = ".6"
    assert jh.dna_to_hp(strategy_hp, dna) == {'hp1': 0.08518987341772151, 'hp2': 3}


def test_dump_exception():
    # uses database, which is not existing during testing
    pass


def test_estimate_average_price():
    assert jh.estimate_average_price(100, 7200, 0, 0) == 7200

    with pytest.raises(TypeError):
        jh.estimate_average_price(100, 7200, 0, None)
        jh.estimate_average_price(100, 7200, None, 0)
        jh.estimate_average_price(100, None, 0, 0)
        jh.estimate_average_price(None, 7200, 0, 0)


def test_estimate_PNL():
    # profit
    assert jh.estimate_PNL(2, 50, 60, 'long') == 20
    assert jh.estimate_PNL(2, 60, 50, 'short') == 20

    # loss
    assert jh.estimate_PNL(2, 50, 60, 'short') == -20
    assert jh.estimate_PNL(2, 60, 50, 'long') == -20

    # profit with fee
    assert jh.estimate_PNL(1, 10, 20, 'long', 0.002) == 9.94
    # loss with fee
    assert jh.estimate_PNL(1, 10, 20, 'short', 0.002) == -10.06

    with pytest.raises(TypeError):
        jh.estimate_PNL(1, 200, 220, 1)
        jh.estimate_PNL(1, 200, 'invalid_input', 'short')
        jh.estimate_PNL(1, 'invalid_input', 220, 'short')
        jh.estimate_PNL('invalid_input', 200, 220, 'short')


def test_estimate_PNL_percentage():
    # profit
    assert jh.estimate_PNL_percentage(1, 200, 220, 'long') == 10
    assert jh.estimate_PNL_percentage(1, 200, 180, 'short') == 10

    # loss
    assert jh.estimate_PNL_percentage(1, 200, 180, 'long') == -10
    assert jh.estimate_PNL_percentage(1, 200, 220, 'short') == -10

    with pytest.raises(TypeError):
        jh.estimate_PNL_percentage(1, 200, 220, 1)
        jh.estimate_PNL_percentage(1, 200, 'invalid_input', 'short')
        jh.estimate_PNL_percentage(1, 'invalid_input', 220, 'short')
        jh.estimate_PNL_percentage('invalid_input', 200, 220, 'short')


def test_file_exists():
    assert jh.file_exists('tests/test_helpers.py') is True


def test_floor_with_precision():
    assert jh.floor_with_precision(1.123) == 1
    assert jh.floor_with_precision(1.123, 1) == 1.1
    assert jh.floor_with_precision(1.123, 2) == 1.12
    assert jh.floor_with_precision(1.123, 3) == 1.123
    assert jh.floor_with_precision(1.123, 4) == 1.123


def test_format_currency():
    assert jh.format_currency(100_000_000) == '100,000,000'
    assert jh.format_currency(100_000_000.23) == '100,000,000.23'


def test_generate_unique_id():
    assert jh.is_valid_uuid(jh.generate_unique_id()) is True
    assert jh.is_valid_uuid('asdfasdfasdfasfsadfsd') is False


def test_get_candle_source():
    candle = np.array(([1575547200000, 146.51, 147.03, 149.02, 146.51, 64788.46651],
                       [1553817660000, 4092.56783507, 4092.5, 4092.56783507, 4092.5, 9.0847059]))
    close = jh.get_candle_source(candle, source_type="close")
    assert close[-1] == 4092.5
    high = jh.get_candle_source(candle, source_type="high")
    assert high[-1] == 4092.56783507
    low = jh.get_candle_source(candle, source_type="low")
    assert low[-1] == 4092.5
    open = jh.get_candle_source(candle, source_type="open")
    assert open[-1] == 4092.56783507
    volume = jh.get_candle_source(candle, source_type="volume")
    assert volume[-1] == 9.0847059
    hl2 = jh.get_candle_source(candle, source_type="hl2")
    assert hl2[-1] == 4092.533917535
    hlc3 = jh.get_candle_source(candle, source_type="hlc3")
    assert hlc3[-1] == 4092.52261169
    ohlc4 = jh.get_candle_source(candle, source_type="ohlc4")
    assert ohlc4[-1] == 4092.533917535


def test_get_config(monkeypatch):
    # assert when config does NOT exist (must return passed default)
    assert jh.get_config('aaaaaaa', 2020) == 2020
    # assert when config does exist
    assert jh.get_config('env.logging.order_submission', 2020) is True
    # assert env is took
    monkeypatch.setenv("ENV_DATABASES_POSTGRES_HOST", "db")
    assert jh.get_config('env.databases.postgres_host', 'default') == 'db'
    monkeypatch.delenv("ENV_DATABASES_POSTGRES_HOST")
    # assert env is took with space
    monkeypatch.setenv("ENV_EXCHANGES_BINANCE_FUTURES_SETTLEMENT_CURRENCY", 'BUSD')
    assert jh.get_config('env.exchanges.Binance Futures.settlement_currency', 'USDT') == 'BUSD'
    monkeypatch.delenv("ENV_EXCHANGES_BINANCE_FUTURES_SETTLEMENT_CURRENCY")


def test_get_strategy_class():
    from jesse.strategies import Strategy
    assert issubclass(jh.get_strategy_class("Test01"), Strategy)


def test_insecure_hash():
    assert jh.insecure_hash("test") == "098f6bcd4621d373cade4e832627b4f6"


def test_insert_list():
    my_list = [0, 1, 2, 3]

    assert jh.insert_list(2, 22, my_list) == [0, 1, 22, 2, 3]
    assert jh.insert_list(0, 22, my_list) == [22, 0, 1, 2, 3]
    assert jh.insert_list(-1, 22, my_list) == [0, 1, 2, 3, 22]

    # assert list is untouched
    assert my_list == [0, 1, 2, 3]


def test_is_backtesting():
    assert jh.is_backtesting() is True


def test_is_collecting_data():
    assert jh.is_collecting_data() is False


def test_is_debuggable():
    debug_item = 'order_submission'
    assert jh.is_debuggable(debug_item) is False


def test_is_debugging():
    assert jh.is_debugging() is False


def test_is_importing_candles():
    assert jh.is_importing_candles() is False


def test_is_live():
    assert jh.is_live() is False


def test_is_livetrading():
    assert jh.is_livetrading() is False


def test_is_optimizing():
    assert jh.is_optimizing() is False


def test_is_paper_trading():
    assert jh.is_paper_trading() is False


def test_is_test_driving():
    assert jh.is_test_driving() is False


def test_is_unit_testing():
    assert jh.is_unit_testing() is True


def test_key():
    exchange = "Exchange"
    symbol = "BTC-USD"
    timeframe = "6h"
    assert jh.key(exchange, symbol) == "Exchange-BTC-USD"
    assert jh.key(exchange, symbol, timeframe) == "Exchange-BTC-USD-6h"


def test_max_timeframe():
    assert jh.max_timeframe(['1m', '3m']) == '3m'
    assert jh.max_timeframe(['3m', '5m']) == '5m'
    assert jh.max_timeframe(['15m', '5m']) == '15m'
    assert jh.max_timeframe(['30m', '15m']) == '30m'
    assert jh.max_timeframe(['30m', '1h']) == '1h'
    assert jh.max_timeframe(['1h', '2h']) == '2h'
    assert jh.max_timeframe(['2h', '3h']) == '3h'
    assert jh.max_timeframe(['4h', '3h']) == '4h'
    assert jh.max_timeframe(['6h', '4h']) == '6h'
    assert jh.max_timeframe(['8h', '4h']) == '8h'
    assert jh.max_timeframe(['6h', '1D']) == '1D'


def test_normalize():
    assert jh.normalize(10, 0, 20) == 0.5
    assert jh.normalize(20, 0, 20) == 1
    assert jh.normalize(0, 0, 20) == 0


def test_now_to_timestamp():
    from jesse.store import store
    assert jh.now_to_timestamp() == store.app.time


def test_np_ffill():
    arr = np.array([0, 1, np.nan, np.nan])
    res = jh.np_ffill(arr)
    expected = np.array([0, 1, 1, 1])

    np.equal(res, expected)


def test_np_shift():
    arr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    res = jh.np_shift(arr, -3)
    expected = np.array([4, 5, 6, 7, 8, 9, 0, 0, 0])

    np.equal(res, expected)


def test_opposite_side():
    assert jh.opposite_side('buy') == 'sell'
    assert jh.opposite_side('sell') == 'buy'


def test_opposite_type():
    assert jh.opposite_type('long') == 'short'
    assert jh.opposite_type('short') == 'long'


def test_orderbook_insertion_index_search():
    ascending_arr = [
        [10, 232],
        [11, 33232],
        [12, 233],
        [33, 21323],
        [44, 23123],
        [55, 2321],
        [66, 23213]
    ]

    assert jh.orderbook_insertion_index_search(ascending_arr, [7, 2]) == (False, 0)
    assert jh.orderbook_insertion_index_search(ascending_arr, [2, 2]) == (False, 0)
    assert jh.orderbook_insertion_index_search(ascending_arr, [32, 2]) == (False, 3)
    assert jh.orderbook_insertion_index_search(ascending_arr, [34, 2]) == (False, 4)
    assert jh.orderbook_insertion_index_search(ascending_arr, [1, 2]) == (False, 0)
    assert jh.orderbook_insertion_index_search(ascending_arr, [66, 2]) == (True, 6)
    assert jh.orderbook_insertion_index_search(ascending_arr, [77, 2]) == (False, 7)

    descending_arr = [
        [66, 232],
        [55, 33232],
        [44, 233],
        [33, 21323],
        [2, 23123],
    ]

    assert jh.orderbook_insertion_index_search(descending_arr, [77, 2], ascending=False) == (False, 0)
    assert jh.orderbook_insertion_index_search(descending_arr, [2, 2], ascending=False) == (True, 4)
    assert jh.orderbook_insertion_index_search(descending_arr, [65, 2], ascending=False) == (False, 1)
    assert jh.orderbook_insertion_index_search(descending_arr, [1, 2], ascending=False) == (False, 5)


def test_orderbook_trim_price():
    # bids
    assert jh.orderbook_trim_price(101.12, False, .1) == 101.1
    assert jh.orderbook_trim_price(101.1, False, .1) == 101.1

    assert jh.orderbook_trim_price(10.12, False, .01) == 10.12
    assert jh.orderbook_trim_price(10.1, False, .01) == 10.1
    assert jh.orderbook_trim_price(10.122, False, .01) == 10.12
    assert jh.orderbook_trim_price(1.1223, False, .001) == 1.122

    # asks
    assert jh.orderbook_trim_price(101.12, True, .1) == 101.2
    assert jh.orderbook_trim_price(101.1, True, .1) == 101.1
    assert jh.orderbook_trim_price(10.12, True, .01) == 10.12
    assert jh.orderbook_trim_price(10.122, True, .01) == 10.13
    assert jh.orderbook_trim_price(1.1223, True, .001) == 1.123


def test_prepare_qty():
    assert jh.prepare_qty(10, 'sell') == -10
    assert jh.prepare_qty(-10, 'buy') == 10

    with pytest.raises(ValueError):
        jh.prepare_qty(-10, 'invalid_input')


def test_python_version():
    import sys
    assert jh.python_version() == sys.version_info[:2]


def test_quote_asset():
    assert jh.quote_asset('BTC-USDT') == 'USDT'
    assert jh.quote_asset('DEFI-USDT') == 'USDT'
    assert jh.quote_asset('DEFI-EUR') == 'EUR'


def test_random_str():
    assert len(jh.random_str(10)) == 10


def test_readable_duration():
    assert jh.readable_duration(604312) == "6 days, 23 hours"


def test_relative_to_absolute():
    from pathlib import Path
    assert jh.relative_to_absolute("tests/test_helpers.py") == str(Path(__file__).absolute())


def test_round_price_for_live_mode():
    np.testing.assert_equal(
        jh.round_price_for_live_mode(np.array([0.0003209123456, 0.0004209123456]), 7),
        np.array([0.0003209, 0.0004209])
    )


def test_round_qty_for_live_mode():
    np.testing.assert_equal(
        jh.round_qty_for_live_mode(np.array([100.3209123456, 100.4299123456]), 2),
        np.array([100.32, 100.42])
    )

    np.testing.assert_equal(
        jh.round_qty_for_live_mode(np.array([0]), 1),
        np.array([0.1])
    )

    np.testing.assert_equal(
        jh.round_qty_for_live_mode(np.array([0]), 2),
        np.array([0.01])
    )

    np.testing.assert_equal(
        jh.round_qty_for_live_mode(np.array([0]), 3),
        np.array([0.001])
    )


def test_round_decimals_down():
    assert jh.round_decimals_down(100.329, 2) == 100.32


def test_secure_hash():
    assert jh.secure_hash('test') == "9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08"


def test_should_execute_silently():
    assert jh.should_execute_silently() is True


def test_side_to_type():
    assert jh.side_to_type("buy") == "long"
    assert jh.side_to_type("sell") == "short"

    # make sure title case works as well
    assert jh.side_to_type("Buy") == "long"
    assert jh.side_to_type("Sell") == "short"


def test_string_after_character():
    assert jh.string_after_character('btcusdt@bookTicker', '@') == 'bookTicker'


def test_style():
    assert jh.style('test', 'bold') == "\x1b[1mtest\x1b[0m"
    assert jh.style('test', 'u') == "\x1b[4mtest\x1b[0m"


def test_terminate_app():
    # uses database, which is not existing during testing
    pass


def test_timeframe_to_one_minutes():
    assert jh.timeframe_to_one_minutes('1m') == 1
    assert jh.timeframe_to_one_minutes('3m') == 3
    assert jh.timeframe_to_one_minutes('5m') == 5
    assert jh.timeframe_to_one_minutes('15m') == 15
    assert jh.timeframe_to_one_minutes('30m') == 30
    assert jh.timeframe_to_one_minutes('1h') == 60
    assert jh.timeframe_to_one_minutes('2h') == 60 * 2
    assert jh.timeframe_to_one_minutes('3h') == 60 * 3
    assert jh.timeframe_to_one_minutes('4h') == 60 * 4
    assert jh.timeframe_to_one_minutes('6h') == 60 * 6
    assert jh.timeframe_to_one_minutes('8h') == 60 * 8
    assert jh.timeframe_to_one_minutes('1D') == 60 * 24


def test_timestamp_to_arrow():
    arrow_time = arrow.get('2015-08-01')
    assert jh.timestamp_to_arrow(1438387200000) == arrow_time


def test_timestamp_to_date():
    assert jh.timestamp_to_date(1438387200000) == "2015-08-01"


def test_timestamp_to_time():
    assert jh.timestamp_to_time(1558770180000) == '2019-05-25T07:43:00+00:00'


def test_today_to_timestamp():
    assert jh.today_to_timestamp() == arrow.utcnow().floor('day').int_timestamp * 1000


def test_type_to_side():
    assert jh.type_to_side('long') == 'buy'
    assert jh.type_to_side('short') == 'sell'


def test_unique_list():
    a = [
        ('Binance', 'BTC', '1m'),
        ('Binance', 'BTC', '5m'),
        ('Binance', 'BTC', '15m'),
        ('Binance', 'BTC', '5m'),
        ('Binance', 'BTC', '1m'),
        ('Binance', 'BTC', '15m'),
    ]

    expected = [
        ('Binance', 'BTC', '1m'),
        ('Binance', 'BTC', '5m'),
        ('Binance', 'BTC', '15m'),
    ]

    assert jh.unique_list(a) == expected


def test_closing_side():
    assert jh.closing_side('Long') == 'sell'
    assert jh.closing_side('Short') == 'buy'


def test_merge_dicts():
    client = {
        'extra': {
            'name': 'Saleh',
            'new_key': 12
        },
        'age': 28
    }

    server = {
        'extra': {
            'name': 'Ocean',
            'water': 100
        },
    }

    expected_result = {'age': 28, 'extra': {'name': 'Ocean', 'water': 100, 'new_key': 12}}

    assert expected_result == jh.merge_dicts(client, server)


def test_get_pid():
    assert os.getpid() == jh.get_pid()


def test_convert_to_env_name():
    assert jh.convert_to_env_name('Testnet Binance Futures') == 'TESTNET_BINANCE_FUTURES'
    assert jh.convert_to_env_name('Testnet Binance') == 'TESTNET_BINANCE'



from jesse.strategies import Strategy


# test_reducing_position_size_after_opening
class Test17(Strategy):
    def should_long(self):
        return self.price < 7

    def go_long(self):
        qty = 2

        self.buy = qty, 7
        self.stop_loss = qty, 5
        self.take_profit = qty, 15

    def update_position(self):
        # reduce the size of position for 1 at current price
        if self.price == 10:
            # should work even without resetting take_profit and stop_loss
            self.take_profit = [
                (1, self.price),
                (1, 15)
            ]

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def filters(self):
        return []

    def should_short(self):
        return False



import jesse.helpers as jh
from jesse.enums import sides, order_statuses
from jesse.models import Order
from jesse.enums import order_types
from .utils import set_up


def test_cancel_order():
    set_up()

    order = Order({
        'id': jh.generate_unique_id(),
        'exchange': 'Sandbox',
        'symbol': 'BTC-USDT',
        'type': order_types.LIMIT,
        'price': 129.33,
        'qty': 10.2041,
        'side': sides.BUY,
        'status': order_statuses.ACTIVE,
        'created_at': jh.now_to_timestamp(),
    })

    assert order.is_canceled is False

    order.cancel()

    assert order.is_canceled is True
    assert order.canceled_at == jh.now_to_timestamp()


def test_execute_order():
    set_up()

    order = Order({
        'id': jh.generate_unique_id(),
        'symbol': 'BTC-USDT',
        'exchange': 'Sandbox',
        'type': order_types.LIMIT,
        'price': 129.33,
        'qty': 10.2041,
        'side': sides.BUY,
        'status': order_statuses.ACTIVE,
        'created_at': jh.now_to_timestamp(),
    })

    assert order.is_executed is False
    assert order.executed_at is None

    order.execute()

    assert order.is_executed is True
    assert order.executed_at == jh.now_to_timestamp()



import os
from datetime import datetime, timedelta

import arrow
import pandas as pd
import quantstats as qs

from jesse.config import config
from jesse.store import store
import jesse.helpers as jh


def quantstats_tearsheet(buy_and_hold_returns: pd.Series, study_name: str) -> None:
    daily_returns = pd.Series(store.app.daily_balance).pct_change(1).values

    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    date_index = pd.date_range(start=start_date, periods=len(store.app.daily_balance))

    returns_time_series = pd.Series(daily_returns, index=date_index)

    mode = config['app']['trading_mode']
    modes = {
        'backtest': ['BT', 'Backtest'],
        'livetrade': ['LT', 'LiveTrade'],
        'papertrade': ['PT', 'PaperTrade']
    }

    os.makedirs('./storage/full-reports', exist_ok=True)

    file_path = f'storage/full-reports/{jh.get_session_id()}.html'.replace(":", "-")

    title = f"{modes[mode][1]} → {arrow.utcnow().strftime('%d %b, %Y %H:%M:%S')} → {study_name}"

    try:
        qs.reports.html(returns=returns_time_series, periods_per_year=365, benchmark=buy_and_hold_returns, title=title, output=file_path)
    except IndexError:
        qs.reports.html(returns=returns_time_series, periods_per_year=365, title=title, output=file_path)
    except:
        raise



# from datetime import timedelta
# import quantstats.plots as _plots
# import quantstats.stats as _stats
# import quantstats.utils as _utils
# from quantstats.reports import _get_trading_periods, metrics
# import numpy as np
# def quantstats_api(benchmark: pd.Series) -> dict:
#     click.clear()
#
#     daily_returns = pd.Series(store.app.daily_balance).pct_change(1).values
#
#     start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
#
#     date_index = pd.date_range(start=start_date, periods=len(store.app.daily_balance))
#
#     returns = pd.Series(daily_returns, index=date_index)
#
#     data = {}
#     periods_per_year = 365
#     rf = 0.
#     compounded = True
#
#     win_year, win_half_year = _get_trading_periods()
#
#     # prepare timeseries
#     returns = _utils._prepare_returns(returns)
#
#     date_range = returns.index.strftime('%e %b, %Y')
#
#     data['date_range'] = f'{date_range[0]} - {date_range[-1]}'
#
#     if benchmark is not None:
#         benchmark = _utils._prepare_benchmark(benchmark, returns.index, rf)
#
#     mtrx = metrics(returns=returns, benchmark=benchmark,
#                    rf=rf, display=False, mode='full',
#                    sep=True, internal="True",
#                    compounded=compounded,
#                    periods_per_year=periods_per_year,
#                    prepare_returns=False)[2:]
#
#     mtrx.index.name = 'Metric'
#     data['metrics'] = mtrx
#
#     if benchmark is not None:
#         yoy = _stats.compare(
#             returns, benchmark, "A", compounded=compounded,
#             prepare_returns=False)
#         yoy.columns = ['Benchmark', 'Strategy', 'Multiplier', 'Won']
#         yoy.index.name = 'Year'
#     else:
#         # pct multiplier
#         yoy = pd.DataFrame(
#             _utils.group_returns(returns, returns.index.year) * 100)
#         yoy.columns = ['Return']
#         yoy['Cumulative'] = _utils.group_returns(
#             returns, returns.index.year, True)
#         yoy['Return'] = yoy['Return'].round(2).astype(str) + '%'
#         yoy['Cumulative'] = (yoy['Cumulative'] *
#                              100).round(2).astype(str) + '%'
#         yoy.index.name = 'Year'
#
#     data['yoy'] = yoy
#
#     dd = _stats.to_drawdown_series(returns)
#     dd_info = _stats.drawdown_details(dd).sort_values(
#         by='max drawdown', ascending=True)[:10]
#
#     dd_info = dd_info[['start', 'end', 'max drawdown', 'days']]
#     dd_info.columns = ['Started', 'Recovered', 'Drawdown', 'Days']
#
#     data['dd_info'] = dd_info
#
#     data['monthly_returns'] = _stats.monthly_returns(returns, eoy=False,
#                                                      compounded=compounded) * 100
#
#     plot_returns = _plots.returns(returns, benchmark,
#                                   figsize=(8, 5), subtitle=False,
#                                   show=False, ylabel=False, cumulative=compounded,
#                                   prepare_returns=False)
#
#     # Get the data from the figure
#     data['plot_returns_strategy'] = plot_returns.get_axes()[0].get_lines()[0].get_data()
#
#     if benchmark is not None:
#         data['plot_returns_benchmark'] = plot_returns.get_axes()[0].get_lines()[1].get_data()
#
#     plot_log_returns = _plots.log_returns(returns, benchmark,
#                                           figsize=(8, 4), subtitle=False,
#                                           show=False, ylabel=False, cumulative=compounded,
#                                           prepare_returns=False)
#
#     # Get the data from the figure
#     data['plot_log_returns_strategy'] = plot_log_returns.get_axes()[0].get_lines()[0].get_data()
#
#     if benchmark is not None:
#         data['plot_log_returns_benchmark'] = plot_log_returns.get_axes()[0].get_lines()[1].get_data()
#
#     if benchmark is not None:
#         vol_returns = _plots.returns(returns, benchmark, match_volatility=True, figsize=(8, 4), subtitle=False,
#                                      show=False, ylabel=False, cumulative=compounded,
#                                      prepare_returns=False)
#         # Get the data from the figure
#         data['plot_vol_returns_strategy'] = vol_returns.get_axes()[0].get_lines()[0].get_data()
#         data['plot_vol_returns_benchmark'] = vol_returns.get_axes()[0].get_lines()[1].get_data()
#
#     yearly_returns = _plots.yearly_returns(returns, benchmark, figsize=(8, 4), subtitle=False,
#                                            show=False, ylabel=False, compounded=compounded,
#                                            prepare_returns=False)
#     ax = yearly_returns.gca()
#
#     data['plot_yearly_returns_labels'] = [l.get_text() for l in ax.get_xticklabels()]
#     values = [rect.get_height() for rect in ax.patches]
#     if benchmark is not None:
#         split = [values[i:i + len(data['plot_yearly_returns_labels'])] for i in range(0, len(values), len(data['plot_yearly_returns_labels']))]
#
#         data['plot_yearly_returns_strategy'] = split[1::2][0] # Start at second element, then every other.
#         data['plot_yearly_returns_benchmark'] = split[::2][0] # Start at first element, then every other.
#     else:
#         data['plot_yearly_returns_strategy'] = values
#
#     data['plot_yearly_returns_mean'] = yearly_returns.get_axes()[0].get_lines()[0].get_ydata()[0]
#
#     histogram = _plots.histogram(returns, figsize=(8, 4), subtitle=False,
#                                  show=False, ylabel=False, compounded=compounded,
#                                  prepare_returns=False)
#     # Get the data from the figure
#     data['plot_distribution_kde'] = histogram.get_axes()[0].get_lines()[2].get_data()
#
#     ax = histogram.gca()
#     y_values = [rect.get_height() for rect in ax.patches]
#     x_values = [rect.get_x() for rect in ax.patches]
#
#     data['plot_distribution_mean'] = histogram.get_axes()[0].get_lines()[0].get_xdata()[0]
#     data['plot_distribution_density'] = y_values
#     data['plot_distribution_percentage'] = x_values
#
#     daily_returns = _plots.daily_returns(returns,
#                                          figsize=(8, 3), subtitle=False,
#                                          show=False, ylabel=False,
#                                          prepare_returns=False)
#     data['plot_daily_returns'] = daily_returns.get_axes()[0].get_lines()[0].get_data()
#
#     if benchmark is not None:
#         rolling_beta = _plots.rolling_beta(returns, benchmark,
#                                            figsize=(8, 3), subtitle=False,
#                                            window1=win_half_year, window2=win_year,
#                                            show=False, ylabel=False,
#                                            prepare_returns=False)
#         data['plot_rolling_beta_6m'] = rolling_beta.get_axes()[0].get_lines()[0].get_data()
#         data['plot_rolling_beta_12m'] = rolling_beta.get_axes()[0].get_lines()[1].get_data()
#         data['plot_rolling_beta_mean'] = rolling_beta.get_axes()[0].get_lines()[2].get_ydata()[0]
#
#     rolling_volatility = _plots.rolling_volatility(returns, benchmark,
#                                                    figsize=(8, 3), subtitle=False,
#                                                    show=False, ylabel=False, period=win_half_year,
#                                                    periods_per_year=win_year)
#
#     # Get the data from the figure
#     data['plot_rolling_volatility_strategy'] = rolling_volatility.get_axes()[0].get_lines()[0].get_data()
#
#     if benchmark is not None:
#         data['plot_rolling_volatility_benchmark'] = rolling_volatility.get_axes()[0].get_lines()[1].get_data()
#
#     data['plot_rolling_volatility_mean'] = rolling_volatility.get_axes()[0].get_lines()[2].get_ydata()[0]
#
#     rolling_sharpe = _plots.rolling_sharpe(returns,
#                                            figsize=(8, 3), subtitle=False,
#                                            show=False, ylabel=False, period=win_half_year,
#                                            periods_per_year=win_year)
#
#     data['plot_rolling_sharpe'] = rolling_sharpe.get_axes()[0].get_lines()[0].get_data()
#
#     data['plot_rolling_sharpe_mean'] = rolling_sharpe.get_axes()[0].get_lines()[1].get_ydata()[0]
#
#     rolling_sortino = _plots.rolling_sortino(returns,
#                                              figsize=(8, 3), subtitle=False,
#                                              show=False, ylabel=False, period=win_half_year,
#                                              periods_per_year=win_year)
#
#     data['plot_rolling_sortino'] = rolling_sortino.get_axes()[0].get_lines()[0].get_data()
#
#     data['plot_rolling_sortino_mean'] = rolling_sortino.get_axes()[0].get_lines()[1].get_ydata()[0]
#
#     drawdowns_periods = _plots.drawdowns_periods(returns, figsize=(8, 4), subtitle=False,
#                                                  show=False, ylabel=False, compounded=compounded,
#                                                  prepare_returns=False)
#
#     data['plot_drawdowns_periods'] = drawdowns_periods.get_axes()[0].get_lines()[0].get_data()
#
#     under_water = _plots.drawdown(returns,
#                     figsize=(8, 3), subtitle=False,
#                     show=False, ylabel=False)
#
#     data['plot_under_water'] = under_water.get_axes()[0].get_lines()[0].get_data()
#
#     data['plot_under_water_mean'] = under_water.get_axes()[0].get_lines()[1].get_ydata()[0]
#
#     port = pd.DataFrame(returns.fillna(0))
#     port.columns = ['Daily']
#
#     apply_fnc = _stats.comp if compounded else np.sum
#
#     port['Weekly'] = port['Daily'].resample(
#         'W-MON').apply(apply_fnc)
#     port['Weekly'].ffill(inplace=True)
#
#     port['Monthly'] = port['Daily'].resample(
#         'M').apply(apply_fnc)
#     port['Monthly'].ffill(inplace=True)
#
#     port['Quarterly'] = port['Daily'].resample(
#         'Q').apply(apply_fnc)
#     port['Quarterly'].ffill(inplace=True)
#
#     port['Yearly'] = port['Daily'].resample(
#         'A').apply(apply_fnc)
#
#     data['quantile'] = port
#
#     jh.dump(list(data.keys()))
#     print('\n\n\n\n')
#     jh.dump(data)
#
#     return data
#



import jesse.helpers as jh
from jesse.enums import order_types
from jesse.exchanges.exchange import Exchange
from jesse.models import Order
from jesse.store import store
from typing import Union

class Sandbox(Exchange):
    def __init__(self, name='Sandbox'):
        super().__init__()
        self.name = name

    def market_order(self, symbol: str, qty: float, current_price: float, side: str, role: str, flags: list) -> Order:
        order = Order({
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': self.name,
            'side': side,
            'type': order_types.MARKET,
            'flag': self.get_exec_inst(flags),
            'qty': jh.prepare_qty(qty, side),
            'price': current_price,
            'role': role
        })

        store.orders.add_order(order)

        store.orders.to_execute.append(order)

        return order

    def limit_order(self, symbol: str, qty: float, price: float, side: str, role: str, flags: list) -> Order:
        order = Order({
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': self.name,
            'side': side,
            'type': order_types.LIMIT,
            'flag': self.get_exec_inst(flags),
            'qty': jh.prepare_qty(qty, side),
            'price': price,
            'role': role
        })

        store.orders.add_order(order)

        return order

    def stop_order(self, symbol: str, qty: float, price: float, side: str, role: str, flags: list) -> Order:
        order = Order({
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': self.name,
            'side': side,
            'type': order_types.STOP,
            'flag': self.get_exec_inst(flags),
            'qty': jh.prepare_qty(qty, side),
            'price': price,
            'role': role
        })

        store.orders.add_order(order)

        return order

    def cancel_all_orders(self, symbol: str) -> None:
        orders = filter(lambda o: o.is_new,
                        store.orders.get_orders(self.name, symbol))

        for o in orders:
            o.cancel()

        if not jh.is_unit_testing():
            store.orders.storage[f'{self.name}-{symbol}'].clear()

    def cancel_order(self, symbol: str, order_id: str) -> None:
        store.orders.get_order_by_id(self.name, symbol, order_id).cancel()

    def get_exec_inst(self, flags: list) -> Union[str, None]:
        if flags:
            return flags[0]
        return None

    def _fetch_precisions(self) -> None:
        pass



from typing import List

import numpy as np

import jesse.helpers as jh
from jesse.config import config
from jesse.libs import DynamicNumpyArray
from jesse.models import store_trade_into_db
from jesse.models.Trade import Trade


class TradesState:
    def __init__(self) -> None:
        self.storage = {}
        self.temp_storage = {}

    def init_storage(self) -> None:
        for c in config['app']['considering_candles']:
            key = jh.key(c[0], c[1])
            self.storage[key] = DynamicNumpyArray((60, 6), drop_at=120)
            self.temp_storage[key] = DynamicNumpyArray((100, 4))

    def add_trade(self, trade: np.ndarray, exchange: str, symbol: str) -> None:
        key = jh.key(exchange, symbol)
        if (
            len(self.temp_storage[key])
            and trade[0] - self.temp_storage[key][0][0] >= 1000
        ):
            arr = self.temp_storage[key]
            buy_arr = np.array(list(filter(lambda x: x[3] == 1, arr)))
            sell_arr = np.array(list(filter(lambda x: x[3] == 0, arr)))

            generated = np.array([
                # timestamp
                arr[0][0],
                # price (weighted average)
                (arr[:][:, 1] * arr[:][:, 2]).sum() / arr[:][:, 2].sum(),
                # buy_qty
                0 if not len(buy_arr) else buy_arr[:, 2].sum(),
                # sell_qty
                0 if not len(sell_arr) else sell_arr[:, 2].sum(),
                # buy_count
                len(buy_arr),
                # sell_count
                len(sell_arr)
            ])

            if jh.is_collecting_data():
                store_trade_into_db(exchange, symbol, generated)
            else:
                self.storage[key].append(generated)

            self.temp_storage[key].flush()
        self.temp_storage[key].append(trade)

    def get_trades(self, exchange: str, symbol: str) -> List[Trade]:
        key = jh.key(exchange, symbol)
        return self.storage[key][:]

    def get_current_trade(self, exchange: str, symbol: str) -> Trade:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1]

    def get_past_trade(self, exchange: str, symbol: str, number_of_trades_ago: int) -> Trade:
        if number_of_trades_ago > 120:
            raise ValueError('Max accepted value for number_of_trades_ago is 120')

        number_of_trades_ago = abs(number_of_trades_ago)
        key = jh.key(exchange, symbol)
        return self.storage[key][-1 - number_of_trades_ago]



from abc import ABC, abstractmethod
from time import sleep
from typing import List, Dict

import numpy as np
import pydash

import jesse.helpers as jh
import jesse.services.logger as logger
import jesse.services.selectors as selectors
from jesse import exceptions
from jesse.enums import sides, trade_types, order_roles
from jesse.models import CompletedTrade, Order, Route, FuturesExchange, SpotExchange, Position
from jesse.models.utils import store_completed_trade_into_db, store_order_into_db
from jesse.services import metrics
from jesse.services.broker import Broker
from jesse.store import store
from jesse.services.cache import cached
from jesse.services import notifier


class Strategy(ABC):
    """
    The parent strategy class which every strategy must extend. It is the heart of the framework!
    """

    def __init__(self) -> None:
        self.id = jh.generate_unique_id()
        self.name = None
        self.symbol = None
        self.exchange = None
        self.timeframe = None
        self.hp = None

        self.index = 0
        self.vars = {}

        self.increased_count = 0
        self.reduced_count = 0

        self.buy = None
        self._buy = None
        self.sell = None
        self._sell = None
        self.stop_loss = None
        self._stop_loss = None
        self.take_profit = None
        self._take_profit = None

        self._entry_orders = []
        self._exit_orders = []

        self.trade: CompletedTrade = None
        self.trades_count = 0

        self._is_executing = False
        self._is_initiated = False
        self._is_handling_updated_order = False

        self.position: Position = None
        self.broker = None

        self._cached_methods = {}
        self._cached_metrics = {}

    def _init_objects(self) -> None:
        """
        This method gets called after right creating the Strategy object. It
        is just a workaround as a part of not being able to set them inside
        self.__init__() for the purpose of removing __init__() methods from strategies.
        """
        self.position = selectors.get_position(self.exchange, self.symbol)
        self.broker = Broker(self.position, self.exchange, self.symbol, self.timeframe)

        if self.hp is None and len(self.hyperparameters()) > 0:
            self.hp = {}
            for dna in self.hyperparameters():
                self.hp[dna['name']] = dna['default']

    @property
    def _price_precision(self) -> int:
        """
        used when live trading because few exchanges require numbers to have a specific precision
        """
        return selectors.get_exchange(self.exchange).vars['precisions'][self.symbol]['price_precision']

    @property
    def _qty_precision(self) -> int:
        """
        used when live trading because few exchanges require numbers to have a specific precision
        """
        return selectors.get_exchange(self.exchange).vars['precisions'][self.symbol]['qty_precision']

    def _broadcast(self, msg: str) -> None:
        """Broadcasts the event to all OTHER strategies

        Arguments:
            msg {str} -- [the message to broadcast]
        """
        from jesse.routes import router

        for r in router.routes:
            # skip self
            if r.strategy.id == self.id:
                continue

            if msg == 'route-open-position':
                r.strategy.on_route_open_position(self)
            elif msg == 'route-close-position':
                r.strategy.on_route_close_position(self)
            elif msg == 'route-increased-position':
                r.strategy.on_route_increased_position(self)
            elif msg == 'route-reduced-position':
                r.strategy.on_route_reduced_position(self)
            elif msg == 'route-canceled':
                r.strategy.on_route_canceled(self)

            r.strategy._detect_and_handle_entry_and_exit_modifications()

    def _on_updated_position(self, order: Order) -> None:
        """
        Handles the after-effect of the executed order

        Note that it assumes that the position has already been affected
        by the executed order.

        Arguments:
            order {Order} -- the executed order object
        """
        # in live-mode, sometimes order-update effects and new execution has overlaps, so:
        self._is_handling_updated_order = True

        role = order.role

        # if the order's role is CLOSE_POSITION but the position qty is not the same as this order's qty,
        # then it's increase_position order (because the position was already open before this)
        if self.trade and role == order_roles.OPEN_POSITION and abs(self.position.qty) != abs(order.qty):
            order.role = order_roles.INCREASE_POSITION
            role = order_roles.INCREASE_POSITION

        # if the order's role is CLOSE_POSITION but the position is still open, then it's reduce_position order
        if role == order_roles.CLOSE_POSITION and self.position.is_open:
            order.role = order_roles.REDUCE_POSITION
            role = order_roles.REDUCE_POSITION

        self._log_position_update(order, role)

        if role == order_roles.OPEN_POSITION:
            self._on_open_position(order)
        elif role == order_roles.CLOSE_POSITION:
            self._on_close_position(order)
        elif role == order_roles.INCREASE_POSITION:
            self._on_increased_position(order)
        elif role == order_roles.REDUCE_POSITION:
            self._on_reduced_position(order)

        self._is_handling_updated_order = False

    def filters(self) -> list:
        return []

    def hyperparameters(self) -> list:
        return []

    def dna(self) -> str:
        return ''

    def _execute_long(self) -> None:
        self.go_long()

        # validation
        if self.buy is None:
            raise exceptions.InvalidStrategy('You forgot to set self.buy. example [qty, price]')
        elif type(self.buy) not in [tuple, list]:
            raise exceptions.InvalidStrategy('self.buy must be either a list or a tuple. example: [qty, price]')

        self._prepare_buy()

        if self.take_profit is not None:
            # validate
            self._validate_take_profit()

            self._prepare_take_profit()

        if self.stop_loss is not None:
            # validate
            self._validate_stop_loss()

            self._prepare_stop_loss()

        # filters
        passed = self._execute_filters()
        if not passed:
            return

        for o in self._buy:
            # MARKET order
            if abs(o[1] - self.price) < 0.0001:
                submitted_order = self.broker.buy_at_market(o[0], order_roles.OPEN_POSITION)
            # STOP order
            elif o[1] > self.price:
                submitted_order = self.broker.start_profit_at(sides.BUY, o[0], o[1], order_roles.OPEN_POSITION)
            # LIMIT order
            elif o[1] < self.price:
                submitted_order = self.broker.buy_at(o[0], o[1], order_roles.OPEN_POSITION)
            else:
                raise ValueError(f'Invalid order price: o[1]:{o[1]}, self.price:{self.price}')

            if submitted_order:
                self._entry_orders.append(submitted_order)

    def _execute_short(self) -> None:
        self.go_short()

        # validation
        if self.sell is None:
            raise exceptions.InvalidStrategy('You forgot to set self.sell. example [qty, price]')
        elif type(self.sell) not in [tuple, list]:
            raise exceptions.InvalidStrategy('self.sell must be either a list or a tuple. example: [qty, price]')

        self._prepare_sell()

        if self.take_profit is not None:
            self._validate_take_profit()
            self._prepare_take_profit()

        if self.stop_loss is not None:
            self._validate_stop_loss()
            self._prepare_stop_loss()

        # filters
        passed = self._execute_filters()
        if not passed:
            return

        for o in self._sell:
            # MARKET order
            if abs(o[1] - self.price) < 0.0001:
                submitted_order = self.broker.sell_at_market(o[0], order_roles.OPEN_POSITION)
            # STOP order
            elif o[1] < self.price:
                submitted_order = self.broker.start_profit_at(sides.SELL, o[0], o[1], order_roles.OPEN_POSITION)
            # LIMIT order
            elif o[1] > self.price:
                submitted_order = self.broker.sell_at(o[0], o[1], order_roles.OPEN_POSITION)
            else:
                raise ValueError(f'Invalid order price: o[1]:{o[1]}, self.price:{self.price}')

            if submitted_order:
                self._entry_orders.append(submitted_order)

    def _prepare_buy(self, make_copies: bool = True) -> None:
        if type(self.buy) is np.ndarray:
            return

        if self.buy is None or self.buy == []:
            # just to make sure we also support None
            self.buy = []
            return

        # create a copy in the placeholders variables so we can detect future modifications
        # also, make it list of orders even if there's only one, to make it easier to loop
        if type(self.buy[0]) not in [list, tuple]:
            self.buy = [self.buy]
        self.buy = self._convert_to_numpy_array(self.buy, 'self.buy')

        if make_copies:
            self._buy = self.buy.copy()

    def _prepare_sell(self, make_copies: bool = True) -> None:
        if type(self.sell) is np.ndarray:
            return

        if self.sell is None or self.sell == []:
            # just to make sure we also support None
            self.sell = []
            return

        # create a copy in the placeholders variables so we can detect future modifications
        # also, make it list of orders even if there's only one, to make it easier to loop
        if type(self.sell[0]) not in [list, tuple]:
            self.sell = [self.sell]
        self.sell = self._convert_to_numpy_array(self.sell, 'self.sell')

        if make_copies:
            self._sell = self.sell.copy()

    def _prepare_stop_loss(self, make_copies: bool = True) -> None:
        # if it's numpy, then it has already been prepared
        if type(self.stop_loss) is np.ndarray:
            return

        if type(self.stop_loss[0]) not in [list, tuple, np.ndarray]:
            self.stop_loss = [self.stop_loss]
        self.stop_loss = self._convert_to_numpy_array(self.stop_loss, 'self.stop_loss')

        if make_copies:
            self._stop_loss = self.stop_loss.copy()

    def _prepare_take_profit(self, make_copies: bool = True) -> None:
        # if it's numpy, then it has already been prepared
        if type(self.take_profit) is np.ndarray:
            return

        if type(self.take_profit[0]) not in [list, tuple, np.ndarray]:
            self.take_profit = [self.take_profit]
        self.take_profit = self._convert_to_numpy_array(self.take_profit, 'self.take_profit')

        if make_copies:
            self._take_profit = self.take_profit.copy()

    def _convert_to_numpy_array(self, arr, name) -> np.ndarray:
        if type(arr) is np.ndarray:
            return arr

        try:
            # create numpy array from list
            arr = np.array(arr, dtype=float)

            if jh.is_live():
                # in livetrade mode, we'll need them rounded
                current_exchange = selectors.get_exchange(self.exchange)

                # skip rounding if the exchange doesn't have values for 'precisions'
                if 'precisions' not in current_exchange.vars:
                    return arr

                price_precision = current_exchange.vars['precisions'][self.symbol]['price_precision']
                qty_precision = current_exchange.vars['precisions'][self.symbol]['qty_precision']

                prices = jh.round_price_for_live_mode(arr[:, 1], price_precision)
                qtys = jh.round_qty_for_live_mode(arr[:, 0], qty_precision)

                arr[:, 0] = qtys
                arr[:, 1] = prices

            return arr
        except ValueError:
            raise exceptions.InvalidShape(
                f'The format of {name} is invalid. \n'
                f'It must be (qty, price) or [(qty, price), (qty, price)] for multiple points; but {arr} was given'
            )

    def _validate_stop_loss(self) -> None:
        if self.stop_loss is None:
            raise exceptions.InvalidStrategy('You forgot to set self.stop_loss. example [qty, price]')
        elif type(self.stop_loss) not in [tuple, list, np.ndarray]:
            raise exceptions.InvalidStrategy('self.stop_loss must be either a list or a tuple. example: [qty, price]')

    def _validate_take_profit(self) -> None:
        if self.take_profit is None:
            raise exceptions.InvalidStrategy('You forgot to set self.take_profit. example [qty, price]')
        elif type(self.take_profit) not in [tuple, list, np.ndarray]:
            raise exceptions.InvalidStrategy('self.take_profit must be either a list or a tuple. example: [qty, price]')

    def _execute_filters(self) -> bool:
        for f in self.filters():
            try:
                passed = f()
            except TypeError:
                raise exceptions.InvalidStrategy(
                    "Invalid filter format. You need to pass filter methods WITHOUT calling them "
                    "(no parentheses must be present at the end)"
                    "\n\n"
                    "\u274C " + "Incorrect Example:\n"
                                "return [\n"
                                "    self.filter_1()\n"
                                "]\n\n"
                                "\u2705 " + "Correct Example:\n"
                                            "return [\n"
                                            "    self.filter_1\n"
                                            "]\n"
                )

            if not passed:
                logger.info(f.__name__)
                self._reset()
                return False

        return True

    @abstractmethod
    def go_long(self) -> None:
        pass

    @abstractmethod
    def go_short(self) -> None:
        pass

    def _execute_cancel(self) -> None:
        """
        cancels everything so that the strategy can keep looking for new trades.
        """
        # validation
        if self.position.is_open:
            raise Exception('cannot cancel orders when position is still open. there must be a bug somewhere.')

        logger.info('cancel all remaining orders to prepare for a fresh start...')

        self.broker.cancel_all_orders()

        self._reset()

        self._broadcast('route-canceled')

        self.on_cancel()

        if not jh.is_unit_testing() and not jh.is_live():
            store.orders.storage[f'{self.exchange}-{self.symbol}'].clear()

    def _reset(self) -> None:
        self.buy = None
        self._buy = None
        self.sell = None
        self._sell = None
        self.stop_loss = None
        self._stop_loss = None
        self.take_profit = None
        self._take_profit = None

        self._entry_orders = []
        self._close_positi = []

        self.increased_count = 0
        self.reduced_count = 0

    def on_cancel(self) -> None:
        """
        what should happen after all active orders have been cancelled
        """
        pass

    @abstractmethod
    def should_long(self) -> bool:
        """are all filters good to execute buy"""
        pass

    @abstractmethod
    def should_short(self) -> bool:
        """are all filters good to execute sell"""
        pass

    @abstractmethod
    def should_cancel(self) -> bool:
        pass

    def before(self) -> None:
        """
        Get's executed BEFORE executing the strategy's logic
        """
        pass

    def after(self) -> None:
        """
        Get's executed AFTER executing the strategy's logic
        """
        pass

    def _update_position(self) -> None:
        self.update_position()

        self._detect_and_handle_entry_and_exit_modifications()

    def _detect_and_handle_entry_and_exit_modifications(self) -> None:
        if self.position.is_close:
            return

        try:
            if self.is_long:
                # prepare format
                self._prepare_buy(make_copies=False)

                # if entry has been modified
                if not np.array_equal(self.buy, self._buy):
                    self._buy = self.buy.copy()

                    # cancel orders
                    for o in self._entry_orders:
                        if o.is_active or o.is_queued:
                            self.broker.cancel_order(o.id)
                    self._entry_orders = [o for o in self._entry_orders if o.is_executed]
                    for o in self._buy:
                        # MARKET order
                        if abs(o[1] - self.price) < 0.0001:
                            submitted_order = self.broker.buy_at_market(o[0], order_roles.OPEN_POSITION)
                        # STOP order
                        elif o[1] > self.price:
                            submitted_order = self.broker.start_profit_at(sides.BUY, o[0], o[1],
                                                                          order_roles.OPEN_POSITION)
                        # LIMIT order
                        elif o[1] < self.price:
                            submitted_order = self.broker.buy_at(o[0], o[1], order_roles.OPEN_POSITION)
                        else:
                            raise ValueError(f'Invalid order price: o[1]:{o[1]}, self.price:{self.price}')

                        if submitted_order:
                            self._entry_orders.append(submitted_order)

            elif self.is_short:
                # prepare format
                self._prepare_sell(make_copies=False)

                # if entry has been modified
                if not np.array_equal(self.sell, self._sell):
                    self._sell = self.sell.copy()

                    # cancel orders
                    for o in self._entry_orders:
                        if o.is_active or o.is_queued:
                            self.broker.cancel_order(o.id)
                    self._entry_orders = [o for o in self._entry_orders if o.is_executed]
                    for o in self._sell:
                        # MARKET order
                        if abs(o[1] - self.price) < 0.0001:
                            submitted_order = self.broker.sell_at_market(o[0], order_roles.OPEN_POSITION)
                        # STOP order
                        elif o[1] < self.price:
                            submitted_order = self.broker.start_profit_at(sides.SELL, o[0], o[1],
                                                                          order_roles.OPEN_POSITION)
                        # LIMIT order
                        elif o[1] > self.price:
                            submitted_order = self.broker.sell_at(o[0], o[1], order_roles.OPEN_POSITION)
                        else:
                            raise ValueError(f'Invalid order price: o[1]:{o[1]}, self.price:{self.price}')

                        if submitted_order:
                            self._entry_orders.append(submitted_order)

            if self.position.is_open and self.take_profit is not None:
                self._validate_take_profit()
                self._prepare_take_profit(False)

                # if _take_profit has been modified
                if not np.array_equal(self.take_profit, self._take_profit):
                    self._take_profit = self.take_profit.copy()

                    # cancel orders
                    for o in self._exit_orders:
                        if o.submitted_via == 'take-profit' and o.is_active or o.is_queued:
                            self.broker.cancel_order(o.id)
                    # clean orders array but leave executed ones
                    self._exit_orders = [o for o in self._exit_orders if o.is_executed]
                    for o in self._take_profit:
                        submitted_order: Order = self.broker.reduce_position_at(
                            o[0],
                            o[1],
                            order_roles.CLOSE_POSITION
                        )
                        if submitted_order:
                            submitted_order.submitted_via = 'take-profit'
                            self._exit_orders.append(submitted_order)

            if self.position.is_open and self.stop_loss is not None:
                self._validate_stop_loss()
                self._prepare_stop_loss(False)

                # if stop_loss has been modified
                if not np.array_equal(self.stop_loss, self._stop_loss):
                    # prepare format
                    self._stop_loss = self.stop_loss.copy()

                    # cancel orders
                    for o in self._exit_orders:
                        if o.submitted_via == 'stop-loss' and o.is_active or o.is_queued:
                            self.broker.cancel_order(o.id)
                    # clean orders array but leave executed ones
                    self._exit_orders = [o for o in self._exit_orders if o.is_executed]
                    for o in self._stop_loss:
                        submitted_order: Order = self.broker.reduce_position_at(
                            o[0],
                            o[1],
                            order_roles.CLOSE_POSITION
                        )
                        if submitted_order:
                            submitted_order.submitted_via = 'stop-loss'
                            self._exit_orders.append(submitted_order)
        except TypeError:
            raise exceptions.InvalidStrategy(
                'Something odd is going on within your strategy causing a TypeError exception. '
                'Try running it with "--debug" in a backtest to see what was going on near the end, and fix it.'
            )
        except:
            raise

        # validations: stop-loss and take-profit should not be the same
        if (
                self.position.is_open
                and (self.stop_loss is not None and self.take_profit is not None)
                and np.array_equal(self.stop_loss, self.take_profit)
        ):
            raise exceptions.InvalidStrategy(
                'stop-loss and take-profit should not be exactly the same. Just use either one of them and it will do.')

    def update_position(self) -> None:
        pass

    def _check(self) -> None:
        """
        Based on the newly updated info, check if we should take action or not
        """
        if not self._is_initiated:
            self._is_initiated = True

        if self._is_handling_updated_order:
            logger.info(
                "Stopped strategy execution at this time because of we're still handling the result "
                "of an order update. Trying again in 3 seconds..."
            )
            sleep(3)

        if jh.is_live() and jh.is_debugging():
            logger.info(f'Executing  {self.name}-{self.exchange}-{self.symbol}-{self.timeframe}')

        # for caution to make sure testing on livetrade won't bleed your account
        if jh.is_test_driving() and store.completed_trades.count >= 2:
            logger.info('Maximum allowed trades in test-drive mode is reached')
            return

        if len(self._entry_orders) and self.is_close and self.should_cancel():
            self._execute_cancel()

            # make sure order cancellation response is received via WS
            if jh.is_live():
                # sleep a little until cancel is received via WS
                sleep(0.1)
                # just in case, sleep some more if necessary
                for _ in range(20):
                    if store.orders.count_active_orders(self.exchange, self.symbol) == 0:
                        break

                    logger.info('sleeping 0.2 more seconds until cancellation is over...')
                    sleep(0.2)

                # If it's still not cancelled, something is wrong. Handle cancellation failure
                if store.orders.count_active_orders(self.exchange, self.symbol) != 0:
                    raise exceptions.ExchangeNotResponding(
                        'The exchange did not respond as expected for order cancellation'
                    )

        if self.position.is_open:
            self._update_position()

        if jh.is_backtesting() or jh.is_unit_testing():
            store.orders.execute_pending_market_orders()

        if self.position.is_close and self._entry_orders == []:
            should_short = self.should_short()
            should_long = self.should_long()
            # validation
            if should_short and should_long:
                raise exceptions.ConflictingRules(
                    'should_short and should_long should not be true at the same time.'
                )
            if should_long:
                self._execute_long()
            elif should_short:
                self._execute_short()

    def _on_open_position(self, order: Order) -> None:
        self.increased_count = 1

        self._broadcast('route-open-position')

        if self.take_profit is not None:
            for o in self._take_profit:
                # validation: make sure take-profit will exit with profit, if not, close the position
                if self.is_long and o[1] <= self.position.entry_price:
                    submitted_order: Order = self.broker.sell_at_market(o[0], order_roles.CLOSE_POSITION)
                    logger.info(
                        'The take-profit is below entry-price for long position, so it will be replaced with a market order instead')
                elif self.is_short and o[1] >= self.position.entry_price:
                    submitted_order: Order = self.broker.buy_at_market(o[0], order_roles.CLOSE_POSITION)
                    logger.info(
                        'The take-profit is above entry-price for a short position, so it will be replaced with a market order instead')
                else:
                    submitted_order: Order = self.broker.reduce_position_at(
                        o[0],
                        o[1],
                        order_roles.CLOSE_POSITION
                    )

                if submitted_order:
                    submitted_order.submitted_via = 'take-profit'
                    self._exit_orders.append(submitted_order)

        if self.stop_loss is not None:
            for o in self._stop_loss:
                # validation: make sure stop-loss will exit with profit, if not, close the position
                if self.is_long and o[1] >= self.position.entry_price:
                    submitted_order: Order = self.broker.sell_at_market(o[0], order_roles.CLOSE_POSITION)
                    logger.info('The stop-loss is above entry-price for long position, so it will be replaced with a market order instead')
                elif self.is_short and o[1] <= self.position.entry_price:
                    submitted_order: Order = self.broker.buy_at_market(o[0], order_roles.CLOSE_POSITION)
                    logger.info('The stop-loss is below entry-price for a short position, so it will be replaced with a market order instead')
                else:
                    submitted_order: Order = self.broker.reduce_position_at(
                        o[0],
                        o[1],
                        order_roles.CLOSE_POSITION
                    )

                if submitted_order:
                    submitted_order.submitted_via = 'stop-loss'
                    self._exit_orders.append(submitted_order)

        self.on_open_position(order)
        self._detect_and_handle_entry_and_exit_modifications()

    def on_open_position(self, order) -> None:
        """
        What should happen after the open position order has been executed
        """
        pass

    def on_close_position(self, order) -> None:
        """
        What should happen after the open position order has been executed
        """
        pass

    def _on_close_position(self, order: Order):
        if not jh.should_execute_silently() or jh.is_debugging():
            logger.info("A closing order has been executed")

        self._broadcast('route-close-position')
        self._execute_cancel()
        self.on_close_position(order)

        self._detect_and_handle_entry_and_exit_modifications()

    def _on_increased_position(self, order: Order) -> None:
        self.increased_count += 1

        self._broadcast('route-increased-position')

        self.on_increased_position(order)

        self._detect_and_handle_entry_and_exit_modifications()

    def on_increased_position(self, order) -> None:
        """
        What should happen after the order (if any) increasing the
        size of the position is executed. Overwrite it if needed.
        And leave it be if your strategy doesn't require it
        """
        pass

    def _on_reduced_position(self, order: Order) -> None:
        """
        prepares for on_reduced_position() is implemented by user
        """
        self.reduced_count += 1

        self._broadcast('route-reduced-position')

        self.on_reduced_position(order)

        self._detect_and_handle_entry_and_exit_modifications()

    def on_reduced_position(self, order) -> None:
        """
        What should happen after the order (if any) reducing the size of the position is executed.
        """
        pass

    def on_route_open_position(self, strategy) -> None:
        """used when trading multiple routes that related

        Arguments:
            strategy {Strategy} -- the strategy that has fired (and not listening to) the event
        """
        pass

    def on_route_close_position(self, strategy) -> None:
        """used when trading multiple routes that related

        Arguments:
            strategy {Strategy} -- the strategy that has fired (and not listening to) the event
        """
        pass

    def on_route_increased_position(self, strategy) -> None:
        """used when trading multiple routes that related

        Arguments:
            strategy {Strategy} -- the strategy that has fired (and not listening to) the event
        """
        pass

    def on_route_reduced_position(self, strategy) -> None:
        """used when trading multiple routes that related

        Arguments:
            strategy {Strategy} -- the strategy that has fired (and not listening to) the event
        """
        pass

    def on_route_canceled(self, strategy) -> None:
        """used when trading multiple routes that related

        Arguments:
            strategy {Strategy} -- the strategy that has fired (and not listening to) the event
        """
        pass

    def _execute(self) -> None:
        """
        Handles the execution permission for the strategy.
        """
        # make sure we don't execute this strategy more than once at the same time.
        if self._is_executing is True:
            return

        self._is_executing = True

        self.before()
        self._check()
        self.after()
        self._clear_cached_methods()

        self._is_executing = False
        self.index += 1

    def _terminate(self) -> None:
        """
        Optional for executing code after completion of a backTest.
        This block will not execute in live use as a live
        Jesse is never ending.
        """
        if not jh.should_execute_silently() or jh.is_debugging():
            logger.info(f"Terminating {self.symbol}...")

        self.terminate()

        self._detect_and_handle_entry_and_exit_modifications()

        # fake execution of market orders in backtest simulation
        if not jh.is_live():
            store.orders.execute_pending_market_orders()

        if jh.is_live():
            return

        if self.position.is_open:
            store.app.total_open_trades += 1
            store.app.total_open_pl += self.position.pnl
            logger.info(
                f"Closed open {self.exchange}-{self.symbol} position at {self.position.current_price} with PNL: {round(self.position.pnl, 4)}({round(self.position.pnl_percentage, 2)}%) because we reached the end of the backtest session."
            )
            # fake a closing (market) order so that the calculations would be correct
            self.broker.reduce_position_at(
                self.position.qty, self.position.current_price, order_roles.CLOSE_POSITION
            )
            return

        if len(self._entry_orders):
            self._execute_cancel()
            logger.info('Canceled open-position orders because we reached the end of the backtest session.')

    def terminate(self):
        pass

    def watch_list(self) -> list:
        """
        returns an array containing an array of key-value items that should
        be logged when backTested, and monitored while liveTraded

        Returns:
            [array[{"key": v, "value": v}]] -- an array of dictionary objects
        """
        return []

    def _clear_cached_methods(self) -> None:
        for m in self._cached_methods.values():
            m.cache_clear()

    @property
    def current_candle(self) -> np.ndarray:
        """
        Returns current trading candle

        :return: np.ndarray
        """
        return store.candles.get_current_candle(self.exchange, self.symbol, self.timeframe).copy()

    @property
    def open(self) -> float:
        """
        Returns the closing price of the current candle for this strategy.
        Just as a helper to use when writing super simple strategies.
        Returns:
            [float] -- the current trading candle's OPEN price
        """
        return self.current_candle[1]

    @property
    def close(self) -> float:
        """
        Returns the closing price of the current candle for this strategy.
        Just as a helper to use when writing super simple strategies.
        Returns:
            [float] -- the current trading candle's CLOSE price
        """
        return self.current_candle[2]

    @property
    def price(self) -> float:
        """
        Same as self.close, except in livetrde, this is rounded as the exchanges require it.

        Returns:
            [float] -- the current trading candle's current(close) price
        """
        return self.position.current_price

    @property
    def high(self) -> float:
        """
        Returns the closing price of the current candle for this strategy.
        Just as a helper to use when writing super simple strategies.
        Returns:
            [float] -- the current trading candle's HIGH price
        """
        return self.current_candle[3]

    @property
    def low(self) -> float:
        """
        Returns the closing price of the current candle for this strategy.
        Just as a helper to use when writing super simple strategies.
        Returns:
            [float] -- the current trading candle's LOW price
        """
        return self.current_candle[4]

    @property
    def candles(self) -> np.ndarray:
        """
        Returns candles for current trading route

        :return: np.ndarray
        """
        return store.candles.get_candles(self.exchange, self.symbol, self.timeframe)

    def get_candles(self, exchange: str, symbol: str, timeframe: str) -> np.ndarray:
        """
        Get candles by passing exchange, symbol, and timeframe

        :param exchange: str
        :param symbol: str
        :param timeframe: str

        :return: np.ndarray
        """
        return store.candles.get_candles(exchange, symbol, timeframe)

    @property
    def orders(self) -> List[Order]:
        """
        Returns all the orders submitted by for this strategy. Just as a helper
        to use when writing super simple strategies.

        Returns:
            [List[Order]] -- orders submitted by strategy
        """
        return store.orders.get_orders(self.exchange, self.symbol)

    @property
    def trades(self) -> List[CompletedTrade]:
        """
        Returns all the completed trades for this strategy.

        Returns:
         [List[CompletedTrade]] -- completed trades by strategy
        """
        return store.completed_trades.trades

    @property
    def metrics(self) -> dict:
        """
        Returns all the metrics of the strategy.
        """
        if self.trades_count not in self._cached_metrics:
            self._cached_metrics[self.trades_count] = metrics.trades(
                store.completed_trades.trades, store.app.daily_balance, final=False
            )
        return self._cached_metrics[self.trades_count]

    @property
    def time(self) -> int:
        """returns the current time"""
        return store.app.time

    @property
    def balance(self) -> float:
        """alias for self.capital"""
        return self.capital

    @property
    def capital(self) -> float:
        """the current capital in the trading exchange"""
        return self.position.exchange.wallet_balance(self.symbol)

    @property
    def available_margin(self) -> float:
        """Current available margin considering leverage"""
        return self.position.exchange.available_margin(self.symbol)

    @property
    def fee_rate(self) -> float:
        return selectors.get_exchange(self.exchange).fee_rate

    def _log_position_update(self, order: Order, role: str) -> None:
        """
        A log can be either about opening, adding, reducing, or closing the position.

        Arguments:
            order {order} -- the order object
        """
        # set the trade_id for the order if we're in the middle of a trade. Otherwise, it
        # is done at order_roles.OPEN_POSITION
        if self.trade:
            order.trade_id = self.trade.id

        if role == order_roles.OPEN_POSITION:
            self.trade = CompletedTrade()
            self.trade.leverage = self.leverage
            self.trade.orders = [order]
            self.trade.timeframe = self.timeframe
            self.trade.id = jh.generate_unique_id()
            order.trade_id = self.trade.id
            self.trade.strategy_name = self.name
            self.trade.exchange = order.exchange
            self.trade.symbol = order.symbol
            self.trade.type = trade_types.LONG if order.side == sides.BUY else trade_types.SHORT
            self.trade.qty = order.qty
            self.trade.opened_at = jh.now_to_timestamp()
            self.trade.entry_candle_timestamp = self.current_candle[0]
        elif role in [order_roles.INCREASE_POSITION, order_roles.REDUCE_POSITION]:
            self.trade.orders.append(order)
            self.trade.qty += order.qty
        elif role == order_roles.CLOSE_POSITION:
            self.trade.exit_candle_timestamp = self.current_candle[0]
            self.trade.orders.append(order)

            # calculate average entry_price price
            sum_price = 0
            sum_qty = 0
            for trade_order in self.trade.orders:
                if not trade_order.is_executed:
                    continue

                if jh.side_to_type(trade_order.side) != self.trade.type:
                    continue

                sum_qty += abs(trade_order.qty)
                sum_price += abs(trade_order.qty) * trade_order.price
            self.trade.entry_price = sum_price / sum_qty

            # calculate average exit_price
            sum_price = 0
            sum_qty = 0
            for trade_order in self.trade.orders:
                if not trade_order.is_executed:
                    continue

                if jh.side_to_type(trade_order.side) == self.trade.type:
                    continue

                sum_qty += abs(trade_order.qty)
                sum_price += abs(trade_order.qty) * trade_order.price

            self.trade.exit_price = sum_price / sum_qty

            self.trade.closed_at = jh.now_to_timestamp()
            self.trade.qty = pydash.sum_by(
                filter(lambda o: o.side == jh.type_to_side(self.trade.type), self.trade.orders),
                lambda o: abs(o.qty)
            )

            store.completed_trades.add_trade(self.trade)
            if jh.is_livetrading():
                store_completed_trade_into_db(self.trade)
            self.trade = None
            self.trades_count += 1
        if jh.is_livetrading():
            store_order_into_db(order)


    @property
    def is_long(self) -> bool:
        return self.position.type == 'long'

    @property
    def is_short(self) -> bool:
        return self.position.type == 'short'

    @property
    def is_open(self) -> bool:
        return self.position.is_open

    @property
    def is_close(self) -> bool:
        return self.position.is_close

    @property
    def average_stop_loss(self) -> float:
        if self._stop_loss is None:
            raise exceptions.InvalidStrategy('You cannot access self.average_stop_loss before setting self.stop_loss')

        arr = self._stop_loss
        return (np.abs(arr[:, 0] * arr[:, 1])).sum() / np.abs(arr[:, 0]).sum()

    @property
    def average_take_profit(self) -> float:
        if self._take_profit is None:
            raise exceptions.InvalidStrategy(
                'You cannot access self.average_take_profit before setting self.take_profit')

        arr = self._take_profit
        return (np.abs(arr[:, 0] * arr[:, 1])).sum() / np.abs(arr[:, 0]).sum()

    @property
    def average_entry_price(self) -> float:
        if self.is_long:
            arr = self._buy
        elif self.is_short:
            arr = self._sell
        elif self.should_long():
            arr = self._buy
        elif self.should_short():
            arr = self._sell
        else:
            return None

        return (np.abs(arr[:, 0] * arr[:, 1])).sum() / np.abs(arr[:, 0]).sum()

    def liquidate(self) -> None:
        """
        closes open position with a MARKET order
        """
        if self.position.is_close:
            return

        if self.position.pnl > 0:
            self.take_profit = self.position.qty, self.price
        else:
            self.stop_loss = self.position.qty, self.price

    @property
    def shared_vars(self) -> dict:
        return store.vars

    @property
    def routes(self) -> List[Route]:
        from jesse.routes import router
        return router.routes

    @property
    def has_active_entry_orders(self) -> bool:
        return len(self._entry_orders) > 0

    @property
    def leverage(self) -> int:
        if type(self.position.exchange) is SpotExchange:
            return 1
        elif type(self.position.exchange) is FuturesExchange:
            return self.position.exchange.futures_leverage
        else:
            raise ValueError('exchange type not supported!')

    @property
    def mark_price(self) -> float:
        return self.position.mark_price

    @property
    def funding_rate(self) -> float:
        return self.position.funding_rate

    @property
    def next_funding_timestamp(self) -> int:
        return self.position.next_funding_timestamp

    @property
    def liquidation_price(self) -> float:
        return self.position.liquidation_price

    @staticmethod
    def log(msg: str, log_type: str = 'info') -> None:
        msg = str(msg)

        if log_type == 'info':
            logger.info(msg)

            if jh.is_live():
                notifier.notify(msg)
        elif log_type == 'error':
            logger.error(msg)

            if jh.is_live():
                notifier.notify(msg)
                notifier.notify_urgently(msg)
        else:
            raise ValueError(f'log_type should be either "info" or "error". You passed {log_type}')
    
    @property
    def all_positions(self) -> Dict[str, Position]:
        positions_dict = {}
        for r in self.routes:
            positions_dict[r.symbol] = r.strategy.position
        return positions_dict

    @property
    def portfolio_value(self) -> float:
        total_position_values = 0
        for key, p in self.all_positions.items():
            total_position_values += p.pnl
        return (total_position_values + self.capital) * self.leverage



from jesse.strategies import Strategy


# test_leverage_property
class TestLeverageProperty2(Strategy):
    def should_long(self) -> bool:
        if self.index == 0:
            assert self.leverage == 2

        return False

    def should_short(self) -> bool:
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass

    def should_cancel(self):
        return False



from jesse.strategies import Strategy


# test_multiple_routes_can_communicate_with_each_other
class Test03(Strategy):
    def should_long(self):
        return False

    def should_short(self):
        return len(self.candles) == 1

    def go_long(self):
        pass

    def go_short(self):
        qty = 1
        sell_price = self.price + 5

        self.sell = qty, sell_price
        self.stop_loss = qty, sell_price + 10
        self.take_profit = qty, sell_price - 10

    def should_cancel(self):
        return False

    def on_route_open_position(self, strategy):
        """

        :param strategy:
        """
        self._execute_cancel()

    def filters(self):
        return []



import csv
import json
import os

import arrow

from jesse.config import config
from jesse.services.tradingview import tradingview_logs
from jesse.store import store
import jesse.helpers as jh


def store_logs(study_name: str = '', export_json: bool = False, export_tradingview: bool = False, export_csv: bool = False) -> None:
    mode = config['app']['trading_mode']

    now = str(arrow.utcnow())[0:19]
    study_name = jh.get_session_id()
    path = f'storage/json/{study_name}.json'
    trades_json = {'trades': [], 'considering_timeframes': config['app']['considering_timeframes']}
    for t in store.completed_trades.trades:
        trades_json['trades'].append(t.toJSON())

    if export_json:
        os.makedirs('./storage/json', exist_ok=True)
        with open(path, 'w+') as outfile:
            def set_default(obj):
                if isinstance(obj, set):
                    return list(obj)
                raise TypeError

            json.dump(trades_json, outfile, default=set_default)

    # store output for TradingView.com's pine-editor
    if export_tradingview:
        tradingview_logs(study_name)

    # also write a CSV file
    if export_csv:
        path = f'storage/csv/{study_name}.csv'
        os.makedirs('./storage/csv', exist_ok=True)

        with open(path, 'w', newline='') as outfile:
            wr = csv.writer(outfile, quoting=csv.QUOTE_ALL)

            for i, t in enumerate(trades_json['trades']):
                if i == 0:
                    # header of CSV file
                    wr.writerow(t.keys())

                wr.writerow(t.values())



from typing import Union

import numpy as np
import talib
from jesse.indicators.ma import ma
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad

from jesse.helpers import get_candle_source, slice_candles


def bollinger_bands_width(candles: np.ndarray, period: int = 20, devup: float = 2, devdn: float = 2, matype: int = 0,
                    devtype: int = 0,
                    source_type: str = "close",
                    sequential: bool = False) -> Union[float, np.ndarray]:
    """
    BBW - Bollinger Bands Width - Bollinger Bands Bandwidth

    :param candles: np.ndarray
    :param period: int - default: 20
    :param devup: float - default: 2
    :param devdn: float - default: 2
    :param matype: int - default: 0
    :param devtype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: BollingerBands(upperband, middleband, lowerband)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    if devtype == 0:
      dev = talib.STDDEV(source, period)
    elif devtype == 1:
      dev = mean_ad(source, period, sequential=True)
    elif devtype == 2:
      dev = median_ad(source, period, sequential=True)

    middlebands = ma(source, period=period, matype=matype, sequential=True)
    upperbands = middlebands + devup * dev
    lowerbands = middlebands - devdn * dev

    if sequential:
        return (upperbands - lowerbands) / middlebands
    else:
        return (upperbands[-1] - lowerbands[-1]) / middlebands[-1]






from typing import Union
import talib
import numpy as np

from jesse.indicators.ma import ma
from jesse.indicators.mean_ad import mean_ad
from jesse.indicators.median_ad import median_ad

try:
    from numba import njit
except ImportError:
    njit = lambda a : a

from jesse.helpers import get_candle_source, slice_candles


def vlma(candles: np.ndarray, min_period: int = 5, max_period: int = 50, matype: int = 0, devtype: int = 0, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Variable Length Moving Average

    :param candles: np.ndarray
    :param min_period: int - default: 5
    :param max_period: int - default: 50
    :param matype: int - default: 0
    :param devtype: int - default: 0
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    mean = ma(source, period=max_period, matype=matype, sequential=True)

    if devtype == 0:
       stdDev = talib.STDDEV(source, max_period)
    elif devtype == 1:
       stdDev = mean_ad(source, max_period, sequential=True)
    elif devtype == 2:
       stdDev = median_ad(source, max_period, sequential=True)

    a = mean - (1.75 * stdDev)
    b = mean - (0.25 * stdDev)
    c = mean + (0.25 * stdDev)
    d = mean + (1.75 * stdDev)

    res = vlma_fast(source, a, b, c, d , min_period, max_period)

    return res if sequential else res[-1]


@njit
def vlma_fast(source, a, b, c, d, min_period, max_period):
    newseries = np.copy(source)
    period = np.zeros_like(source)
    for i in range(1, source.shape[0]):
        nz_period = period[i - 1] if period[i - 1] != 0 else max_period
        period[i] = nz_period + 1 if b[i] <= source[i] <= c[i] else nz_period - 1 if source[i] < a[i] or source[i] > d[i] else nz_period
        period[i] = max(min(period[i], max_period), min_period)
        sc = 2 / (period[i] + 1)
        newseries[i] = (source[i] * sc) + ((1 - sc) * newseries[i - 1])
    return newseries



from jesse.modes.import_candles_mode.drivers.binance_inverse_futures import encode_symbol, decode_symbol


def test_encode_symbol():
    assert encode_symbol('BTC-PERP') == 'BTCUSD_PERP'
    assert encode_symbol('BTC-200925') == 'BTCUSD_200925'


def test_decode_symbol():
    assert decode_symbol('BTCUSD_PERP') == 'BTC-PERP'
    assert decode_symbol('BTCUSD_200925') == 'BTC-200925'



# import pickle
from abc import ABC
from datetime import timedelta
from random import randint, choices

from timeloop import Timeloop

from jesse import sync_publish
from jesse.services.redis import process_status
from multiprocessing import Process, Manager
import click
import numpy as np
import pydash
import jesse.helpers as jh
import jesse.services.logger as logger
from jesse.store import store
import os
# import json
# from pandas import json_normalize
from jesse import exceptions
from jesse.modes.optimize_mode.fitness import get_and_add_fitness_to_the_bucket, create_baby
from jesse.routes import router
from numpy import ndarray
from multiprocessing import cpu_count
from jesse.services.progressbar import Progressbar


class Optimizer(ABC):
    def __init__(
            self,
            training_candles: ndarray, testing_candles: ndarray,
            optimal_total: int, cpu_cores: int,
            csv: bool,
            export_json: bool,
            start_date: str, finish_date: str,
            charset: str = r'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvw',
            fitness_goal: float = 1,
    ) -> None:
        if len(router.routes) != 1:
            raise NotImplementedError('optimize_mode mode only supports one route at the moment')

        self.strategy_name = router.routes[0].strategy_name
        self.exchange = router.routes[0].exchange
        self.symbol = router.routes[0].symbol
        self.timeframe = router.routes[0].timeframe
        strategy_class = jh.get_strategy_class(self.strategy_name)
        self.strategy_hp = strategy_class.hyperparameters(None)
        solution_len = len(self.strategy_hp)

        if solution_len == 0:
            raise exceptions.InvalidStrategy('Targeted strategy does not implement a valid hyperparameters() method.')

        self.started_index = 0
        self.start_time = jh.now_to_timestamp()
        self.population = []
        self.iterations = 2000 * solution_len
        self.population_size = solution_len * 100
        self.solution_len = solution_len
        self.charset = charset
        self.fitness_goal = fitness_goal
        self.cpu_cores = 0
        self.optimal_total = optimal_total
        self.training_candles = training_candles
        self.testing_candles = testing_candles
        self.average_execution_seconds = 0

        # check for termination event once per second
        tl_0 = Timeloop()
        @tl_0.job(interval=timedelta(seconds=1))
        def check_for_termination():
            if process_status() != 'started':
                raise exceptions.Termination
        tl_0.start()

        options = {
            'strategy_name': self.strategy_name,
            'exchange': self.exchange,
            'symbol': self.symbol,
            'timeframe': self.timeframe,
            'strategy_hp': self.strategy_hp,
            'csv': csv,
            'json': export_json,
            'start_date': start_date,
            'finish_date': finish_date,
        }

        self.options = {} if options is None else options
        os.makedirs('./storage/temp/optimize', exist_ok=True)
        self.temp_path = f"./storage/temp/optimize/{self.options['strategy_name']}-{self.options['exchange']}-{self.options['symbol']}-{self.options['timeframe']}-{self.options['start_date']}-{self.options['finish_date']}.pickle"

        if fitness_goal > 1 or fitness_goal < 0:
            raise ValueError('fitness scores must be between 0 and 1')

        if not optimal_total > 0:
            raise ValueError('optimal_total must be bigger than 0')

        # # if temp file exists, load data to resume previous session
        # if jh.file_exists(self.temp_path) and click.confirm(
        #         'Previous session detected. Do you want to resume?', default=True
        # ):
        #     self.load_progress()

        if cpu_cores > cpu_count():
            raise ValueError(f'Entered cpu cores number is more than available on this machine which is {cpu_count()}')
        elif cpu_cores == 0:
            self.cpu_cores = cpu_count()
        else:
            self.cpu_cores = cpu_cores

    def generate_initial_population(self) -> None:
        """
        generates the initial population
        """
        length = int(self.population_size / self.cpu_cores)

        progressbar = Progressbar(length)
        for i in range(length):
            people = []
            with Manager() as manager:
                dna_bucket = manager.list([])
                workers = []

                try:
                    for _ in range(self.cpu_cores):
                        dna = ''.join(choices(self.charset, k=self.solution_len))
                        w = Process(
                            target=get_and_add_fitness_to_the_bucket,
                            args=(
                                dna_bucket, jh.get_config('env.optimization'), router.formatted_routes, router.formatted_extra_routes,
                                self.strategy_hp, dna, self.training_candles, self.testing_candles,
                                self.optimal_total
                            )
                        )
                        w.start()
                        workers.append(w)

                    # join workers
                    for w in workers:
                        w.join()
                        if w.exitcode > 0:
                            logger.error(f'a process exited with exitcode: {w.exitcode}')
                except exceptions.Termination:
                    self._handle_termination(manager, workers)

                for d in dna_bucket:
                    people.append({
                        'dna': d[0],
                        'fitness': d[1],
                        'training_log': d[2],
                        'testing_log': d[3]
                    })

            # update dashboard
            self.update_progressbar(progressbar)

            # general_info streams
            general_info = {
                'started_at': jh.timestamp_to_arrow(self.start_time).humanize(),
                'index': f'{(i + 1) * self.cpu_cores}/{self.population_size}',
                'errors_info_count': f'{len(store.logs.errors)}/{len(store.logs.info)}',
                'trading_route': f'{router.routes[0].exchange}, {router.routes[0].symbol}, {router.routes[0].timeframe}, {router.routes[0].strategy_name}',
                'average_execution_seconds': self.average_execution_seconds
            }
            if jh.is_debugging():
                general_info['population_size'] = self.population_size
                general_info['iterations'] = self.iterations
                general_info['solution_length'] = self.solution_len
            sync_publish('general_info', general_info)

            for p in people:
                self.population.append(p)

        sync_publish('progressbar', {
            'current': 100,
            'estimated_remaining_seconds': 0
        })
        # sort the population
        self.population = list(sorted(self.population, key=lambda x: x['fitness'], reverse=True))

    def select_person(self) -> dict:
        # len(self.population) instead of self.population_size because some DNAs might not have been created due to errors
        # to fix an issue with being less than 100 population_len (which means there's only on hyperparameter in the strategy)
        population_len = len(self.population)
        random_index = np.random.choice(population_len, int(population_len / 100), replace=False)
        chosen_ones = [self.population[r] for r in random_index]

        return pydash.max_by(chosen_ones, 'fitness')

    def evolve(self) -> list:
        """
        the main method, that runs the evolutionary algorithm
        """
        if self.started_index == 0:
            self.generate_initial_population()

            if len(self.population) < 0.5 * self.population_size:
                raise ValueError('Too many errors! less than half of the expected population size could be generated.')

            # if even our best individual is too weak, then we better not continue
            if self.population[0]['fitness'] == 0.0001:
                raise exceptions.InvalidStrategy(
                    'Cannot continue because no individual with the minimum fitness-score was found. '
                    'Your strategy seems to be flawed or maybe it requires modifications. ')

        loop_length = int(self.iterations / self.cpu_cores)

        i = self.started_index
        progressbar = Progressbar(loop_length)
        while i < loop_length:
            with Manager() as manager:
                people_bucket = manager.list([])
                workers = []

                try:
                    for _ in range(self.cpu_cores):
                        mommy = self.select_person()
                        daddy = self.select_person()
                        w = Process(
                            target=create_baby,
                            args=(
                                people_bucket, mommy, daddy, self.solution_len, self.charset,
                                jh.get_config('env.optimization'), router.formatted_routes,
                                router.formatted_extra_routes,
                                self.strategy_hp, self.training_candles, self.testing_candles,
                                self.optimal_total
                            )
                        )
                        w.start()
                        workers.append(w)

                    for w in workers:
                        w.join()
                        if w.exitcode > 0:
                            logger.error(f'a process exited with exitcode: {w.exitcode}')
                except exceptions.Termination:
                    self._handle_termination(manager, workers)

                # update dashboard
                click.clear()
                self.update_progressbar(progressbar)
                # general_info streams
                general_info = {
                    'started_at': jh.timestamp_to_arrow(self.start_time).humanize(),
                    'index': f'{(i + 1) * self.cpu_cores}/{self.iterations}',
                    'errors_info_count': f'{len(store.logs.errors)}/{len(store.logs.info)}',
                    'trading_route': f'{router.routes[0].exchange}, {router.routes[0].symbol}, {router.routes[0].timeframe}, {router.routes[0].strategy_name}',
                    'average_execution_seconds': self.average_execution_seconds
                }
                if jh.is_debugging():
                    general_info['population_size'] = self.population_size
                    general_info['iterations'] = self.iterations
                    general_info['solution_length'] = self.solution_len
                sync_publish('general_info', general_info)

                if self.population_size > 50:
                    number_of_ind_to_show = 40
                elif self.population_size > 20:
                    number_of_ind_to_show = 15
                elif self.population_size > 9:
                    number_of_ind_to_show = 9
                else:
                    raise ValueError('self.population_size cannot be less than 10')

                best_candidates = [{
                        'rank': j + 1,
                        'dna': self.population[j]['dna'],
                        'fitness': round(self.population[j]['fitness'], 4),
                        'training_win_rate': self.population[j]['training_log']['win-rate'],
                        'training_total_trades': self.population[j]['training_log']['total'],
                        'training_pnl': self.population[j]['training_log']['PNL'],
                        'testing_win_rate': self.population[j]['testing_log']['win-rate'],
                        'testing_total_trades': self.population[j]['testing_log']['total'],
                        'testing_pnl': self.population[j]['testing_log']['PNL'],
                    } for j in range(number_of_ind_to_show)]
                sync_publish('best_candidates', best_candidates)

                # one person has to die and be replaced with the newborn baby
                for baby in people_bucket:
                    # never kill our best performer
                    random_index = randint(1, len(self.population) - 1)
                    self.population[random_index] = baby
                    self.population = list(sorted(self.population, key=lambda x: x['fitness'], reverse=True))

                    # reaching the fitness goal could also end the process
                    if baby['fitness'] >= self.fitness_goal:
                        self.update_progressbar(progressbar, finished=True)
                        sync_publish('alert', {
                            'message': f'Fitness goal reached after iteration {i*self.cpu_cores}',
                            'type': 'success'
                        })
                        return baby

                # TODO: bring back progress resumption
                # # save progress after every n iterations
                # if i != 0 and int(i * self.cpu_cores) % 50 == 0:
                #     self.save_progress(i)

                # TODO: bring back
                # # store a take_snapshot of the fittest individuals of the population
                # if i != 0 and i % int(100 / self.cpu_cores) == 0:
                #     self.take_snapshot(i * self.cpu_cores)

                i += 1

        sync_publish('alert', {
            'message': f"Finished {self.iterations} iterations. Check your best DNA candidates, "
                       f"if you don't like any of them, try modifying your strategy.",
            'type': 'success'
        })
        return self.population

    def run(self) -> list:
        return self.evolve()


    @staticmethod
    def _handle_termination(manager, workers):
        logger.info('Terminating session...')

        # terminate all workers
        for w in workers:
            w.terminate()

        # shutdown the manager process manually since garbage collection cannot won't get to do it for us
        manager.shutdown()

        # now we can terminate the main session safely
        raise exceptions.Termination

    def update_progressbar(self, progressbar, finished=False):
        if finished:
            progressbar.finish()
        else:
            progressbar.update()
        self.average_execution_seconds = progressbar.average_execution_seconds / self.cpu_cores
        sync_publish('progressbar', {
            'current': progressbar.current,
            'estimated_remaining_seconds': progressbar.estimated_remaining_seconds
        })


    # def save_progress(self, iterations_index: int) -> None:
    #     """
    #     pickles data so we can later resume optimizing
    #     """
    #     data = {
    #         'population': self.population,
    #         'iterations': self.iterations,
    #         'iterations_index': iterations_index,
    #         'population_size': self.population_size,
    #         'solution_len': self.solution_len,
    #         'charset': self.charset,
    #         'fitness_goal': self.fitness_goal,
    #         'options': self.options,
    #     }
    #
    #     with open(self.temp_path, 'wb') as f:
    #         pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    # def load_progress(self) -> None:
    #     """
    #     Unpickles data to resume from previous optimizing session population
    #     """
    #     with open(self.temp_path, 'rb') as f:
    #         data = pickle.load(f)
    #
    #     self.population = data['population']
    #     self.iterations = data['iterations']
    #     self.started_index = data['iterations_index']
    #     self.population_size = data['population_size']
    #     self.solution_len = data['solution_len']
    #     self.charset = data['charset']
    #     self.fitness_goal = data['fitness_goal']
    #     self.options = data['options']

    # def take_snapshot(self, index: int) -> None:
    #     """
    #     stores a snapshot of the fittest population members into a file.
    #     """
    #     study_name = f"{self.options['strategy_name']}-{self.options['exchange']}-{self.options['symbol']}-{self.options['timeframe']}-{self.options['start_date']}-{self.options['finish_date']}"
    #
    #     dna_json = {'snapshot': []}
    #     for i in range(30):
    #         dna_json['snapshot'].append(
    #             {'iteration': index, 'dna': self.population[i]['dna'], 'fitness': self.population[i]['fitness'],
    #              'training_log': self.population[i]['training_log'], 'testing_log': self.population[i]['testing_log'],
    #              'parameters': jh.dna_to_hp(self.options['strategy_hp'], self.population[i]['dna'])})
    #
    #     path = f'./storage/genetics/{study_name}.txt'
    #     os.makedirs('./storage/genetics', exist_ok=True)
    #     txt = ''
    #     with open(path, 'a', encoding="utf-8") as f:
    #         txt += '\n\n'
    #         txt += f'# iteration {index}'
    #         txt += '\n'
    #
    #         for i in range(30):
    #             log = f"win-rate: {self.population[i]['training_log']['win-rate']} %, total: {self.population[i]['training_log']['total']}, PNL: {self.population[i]['training_log']['PNL']} % || win-rate: {self.population[i]['testing_log']['win-rate']} %, total: {self.population[i]['testing_log']['total']}, PNL: {self.population[i]['testing_log']['PNL']} %"
    #
    #             txt += '\n'
    #             txt += f"{i + 1} ==  {self.population[i]['dna']}  ==  {self.population[i]['fitness']}  ==  {log}"
    #
    #         f.write(txt)
    #
    #     if self.options['csv']:
    #         path = f'storage/genetics/csv/{study_name}.csv'
    #         os.makedirs('./storage/genetics/csv', exist_ok=True)
    #         exists = os.path.exists(path)
    #
    #         df = json_normalize(dna_json['snapshot'])
    #
    #         with open(path, 'a', newline='', encoding="utf-8") as outfile:
    #             if not exists:
    #                 # header of CSV file
    #                 df.to_csv(outfile, header=True, index=False, encoding='utf-8')
    #
    #             df.to_csv(outfile, header=False, index=False, encoding='utf-8')
    #
    #     if self.options['json']:
    #         path = f'storage/genetics/json/{study_name}.json'
    #         os.makedirs('./storage/genetics/json', exist_ok=True)
    #         exists = os.path.exists(path)
    #
    #         mode = 'r+' if exists else 'w'
    #         with open(path, mode, encoding="utf-8") as file:
    #             if not exists:
    #                 snapshots = {"snapshots": []}
    #                 snapshots["snapshots"].append(dna_json['snapshot'])
    #                 json.dump(snapshots, file, ensure_ascii=False)
    #             else:
    #                 # file exists - append
    #                 file.seek(0)
    #                 data = json.load(file)
    #                 data["snapshots"].append(dna_json['snapshot'])
    #                 file.seek(0)
    #                 json.dump(data, file, ensure_ascii=False)
    #             file.write('\n')



from jesse.strategies import Strategy, cached
import jesse.indicators as ta
from jesse import utils


class ExampleStrategy(Strategy):
    def should_long(self) -> bool:
        return False

    def should_short(self) -> bool:
        return False

    def should_cancel(self) -> bool:
        return True

    def go_long(self):
        pass

    def go_short(self):
        pass



from jesse.strategies import Strategy


# test_modifying_stop_loss_after_part_of_position_is_already_reduced_with_stop_loss
class Test14(Strategy):
    def should_long(self):
        return self.price < 7

    def should_short(self):
        return False

    def go_long(self):
        qty = 1.5
        self.buy = qty, 7
        self.stop_loss = [
            (0.5, 6),
            (0.5, 5),
            (0.5, 4)
        ]
        self.take_profit = qty, 13

    def update_position(self):
        if self.reduced_count > 0:
            self.stop_loss = self.position.qty, 4

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def filters(self):
        return []



from typing import Union

import numpy as np
import talib

from jesse.helpers import slice_candles


def adosc(candles: np.ndarray, fast_period: int = 3, slow_period: int = 10, sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    ADOSC - Chaikin A/D Oscillator

    :param candles: np.ndarray
    :param fast_period: int - default: 3
    :param slow_period: int - default: 10
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    res = talib.ADOSC(candles[:, 3], candles[:, 4], candles[:, 2], candles[:, 5], fastperiod=fast_period,
                      slowperiod=slow_period)

    return res if sequential else res[-1]



from abc import ABC, abstractmethod

from jesse.models import Order


class Exchange(ABC):
    name = ''
    fee_rate = None

    # current holding assets
    assets = {}
    # used for calculating available balance in futures mode:
    temp_reduced_amount = {}
    # current available assets (dynamically changes based on active orders)
    available_assets = {}
    # used for calculating final performance metrics
    starting_assets = {}

    # some exchanges might require even further info
    vars = {}

    def __init__(self, name: str, starting_assets: list, fee_rate: float, exchange_type: str):
        self.name = name
        self.type = exchange_type.lower()

        for item in starting_assets:
            self.assets[item['asset']] = item['balance']
            self.temp_reduced_amount[item['asset']] = 0

        self.starting_assets = self.assets.copy()
        self.available_assets = self.assets.copy()
        self.fee_rate = fee_rate

    @abstractmethod
    def wallet_balance(self, symbol: str = '') -> float:
        pass

    @abstractmethod
    def available_margin(self, symbol: str = '') -> float:
        pass

    @abstractmethod
    def on_order_submission(self, order: Order, skip_market_order: bool = True) -> None:
        pass

    @abstractmethod
    def on_order_execution(self, order: Order) -> None:
        pass

    @abstractmethod
    def on_order_cancellation(self, order: Order) -> None:
        pass

    @abstractmethod
    def add_realized_pnl(self, realized_pnl: float) -> None:
        pass

    @abstractmethod
    def charge_fee(self, amount: float) -> None:
        pass



from random import randint

import jesse.helpers as jh
from jesse.enums import exchanges, sides, order_types, order_statuses
from jesse.models import Order

first_timestamp = 1552309186171


def fake_order(attributes: dict = None) -> Order:
    """

    :param attributes:
    :return:
    """
    if attributes is None:
        attributes = {}

    global first_timestamp
    first_timestamp += 60000
    exchange = exchanges.SANDBOX
    symbol = 'BTC-USD'
    side = sides.BUY
    order_type = order_types.LIMIT
    price = randint(40, 100)
    qty = randint(1, 10)
    status = order_statuses.ACTIVE
    created_at = first_timestamp

    return Order({
        "id": jh.generate_unique_id(),
        'symbol': attributes.get('symbol', symbol),
        'exchange': attributes.get('exchange', exchange),
        'side': attributes.get('side', side),
        'type': attributes.get('type', order_type),
        'qty': attributes.get('qty', qty),
        'price': attributes.get('price', price),
        'status': attributes.get('status', status),
        'created_at': attributes.get('created_at', created_at),
    })



from jesse.factories import range_candles
from jesse.services.candle import *


def test_candle_includes_price():
    c = np.array([1543387200000, 10, 20, 25, 5, 195])

    assert candle_includes_price(c, 5)
    assert candle_includes_price(c, 15)
    assert candle_includes_price(c, 25)

    assert not candle_includes_price(c, 4)
    assert not candle_includes_price(c, 26)


def test_generate_candle_from_one_minutes():
    candles = range_candles(5)

    five_minutes_candle = generate_candle_from_one_minutes('5m', candles)

    assert five_minutes_candle[0] == candles[0][0]
    assert five_minutes_candle[1] == candles[0][1]
    assert five_minutes_candle[2] == candles[-1][2]
    assert five_minutes_candle[3] == candles[:, 3].max()
    assert five_minutes_candle[4] == candles[:, 4].min()
    assert five_minutes_candle[5] == candles[:, 5].sum()


def test_is_bearish():
    c = np.array([1543387200000, 200, 190, 220, 180, 195])
    assert is_bearish(c)


def test_is_bullish():
    c = np.array([1543387200000, 190, 200, 220, 180, 195])
    assert is_bullish(c)


def test_split_candle():
    """
    these values has been tested from my thoughts on paper. You need to reproduce my drawings for them to make sense
    """
    bull = np.array([1111, 10, 20, 25, 5, 2222])
    bear = np.array([1111, 20, 10, 25, 5, 2222])

    # bullish candle, low < price < open
    np.testing.assert_equal(
        split_candle(bull, 7),
        (
            np.array([1111, 10, 7, 10, 7, 2222]),
            np.array([1111, 7, 20, 25, 5, 2222]),
        )
    )
    # bearish candle, open < price < high
    np.testing.assert_equal(
        split_candle(bear, 23),
        (
            np.array([1111, 20, 23, 23, 20, 2222]),
            np.array([1111, 23, 10, 25, 5, 2222]),
        )
    )

    # bullish candle, price == open
    np.testing.assert_equal(
        split_candle(bull, bull[1]),
        (bull, bull)
    )
    # bearish candle, price == open
    np.testing.assert_equal(
        split_candle(bear, bear[1]),
        (bear, bear)
    )

    # bearish candle,  low < price < close
    np.testing.assert_equal(
        split_candle(bear, 7),
        (
            np.array([1111, 20, 7, 25, 7, 2222]),
            np.array([1111, 7, 10, 10, 5, 2222]),
        )
    )
    # bullish candle,  close < price < high
    np.testing.assert_equal(
        split_candle(bull, 23),
        (
            np.array([1111, 10, 23, 23, 5, 2222]),
            np.array([1111, 23, 20, 25, 20, 2222]),
        )
    )

    # bearish candle,  price == close
    np.testing.assert_equal(
        split_candle(bear, 10),
        (
            np.array([1111, 20, 10, 25, 10, 2222]),
            np.array([1111, 10, 10, 10, 5, 2222]),
        )
    )
    # bullish candle,  close < price < high
    np.testing.assert_equal(
        split_candle(bull, 20),
        (
            np.array([1111, 10, 20, 20, 5, 2222]),
            np.array([1111, 20, 20, 25, 20, 2222]),
        )
    )

    # bearish candle,  price == high
    np.testing.assert_equal(
        split_candle(bear, 25),
        (
            np.array([1111, 20, 25, 25, 20, 2222]),
            np.array([1111, 25, 10, 25, 5, 2222]),
        )
    )
    # bullish candle,  price == low
    np.testing.assert_equal(
        split_candle(bull, 5),
        (
            np.array([1111, 10, 5, 10, 5, 2222]),
            np.array([1111, 5, 20, 25, 5, 2222]),
        )
    )

    # bearish candle,  price == low
    np.testing.assert_equal(
        split_candle(bear, 5),
        (
            np.array([1111, 20, 5, 25, 5, 2222]),
            np.array([1111, 5, 10, 10, 5, 2222]),
        )
    )
    # bullish candle,  price == high
    np.testing.assert_equal(
        split_candle(bull, 25),
        (
            np.array([1111, 10, 25, 25, 5, 2222]),
            np.array([1111, 25, 20, 25, 20, 2222]),
        )
    )

    # bearish candle, close < price < open
    np.testing.assert_equal(
        split_candle(bear, 15),
        (
            np.array([1111, 20, 15, 25, 15, 2222]),
            np.array([1111, 15, 10, 15, 5, 2222]),
        )
    )
    # bullish candle, open < price < close
    np.testing.assert_equal(
        split_candle(bull, 15),
        (
            np.array([1111, 10, 15, 15, 5, 2222]),
            np.array([1111, 15, 20, 25, 15, 2222]),
        )
    )



import asyncio
import json
import os
import warnings
from typing import Optional
import click
import pkg_resources
from fastapi import BackgroundTasks, Query, Header
from starlette.websockets import WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse, FileResponse
from fastapi.staticfiles import StaticFiles
from jesse.services import auth as authenticator
from jesse.services.redis import async_redis, async_publish, sync_publish
from jesse.services.web import fastapi_app, BacktestRequestJson, ImportCandlesRequestJson, CancelRequestJson, \
    LoginRequestJson, ConfigRequestJson, LoginJesseTradeRequestJson, NewStrategyRequestJson, FeedbackRequestJson, \
    ReportExceptionRequestJson, OptimizationRequestJson
import uvicorn
from asyncio import Queue
import jesse.helpers as jh
import time

# to silent stupid pandas warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


# variable to know if the live trade plugin is installed
HAS_LIVE_TRADE_PLUGIN = True
try:
    import jesse_live
except ModuleNotFoundError:
    HAS_LIVE_TRADE_PLUGIN = False


def validate_cwd() -> None:
    """
    make sure we're in a Jesse project
    """
    if not jh.is_jesse_project():
        print(
            jh.color(
                'Current directory is not a Jesse project. You must run commands from the root of a Jesse project. Read this page for more info: https://docs.jesse.trade/docs/getting-started/#create-a-new-jesse-project',
                'red'
            )
        )
        os._exit(1)


# print(os.path.dirname(jesse))
JESSE_DIR = os.path.dirname(os.path.realpath(__file__))


# load homepage
@fastapi_app.get("/")
async def index():
    return FileResponse(f"{JESSE_DIR}/static/index.html")


@fastapi_app.post("/terminate-all")
async def terminate_all(authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.multiprocessing import process_manager

    process_manager.flush()
    return JSONResponse({'message': 'terminating all tasks...'})


@fastapi_app.post("/shutdown")
async def shutdown(background_tasks: BackgroundTasks, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    background_tasks.add_task(jh.terminate_app)
    return JSONResponse({'message': 'Shutting down...'})


@fastapi_app.post("/auth")
def auth(json_request: LoginRequestJson):
    return authenticator.password_to_token(json_request.password)


@fastapi_app.post("/make-strategy")
def make_strategy(json_request: NewStrategyRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import strategy_maker
    return strategy_maker.generate(json_request.name)


@fastapi_app.post("/feedback")
def feedback(json_request: FeedbackRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    return jesse_trade.feedback(json_request.description, json_request.email)


@fastapi_app.post("/report-exception")
def report_exception(json_request: ReportExceptionRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services import jesse_trade
    return jesse_trade.report_exception(
        json_request.description,
        json_request.traceback,
        json_request.mode,
        json_request.attach_logs,
        json_request.session_id,
        json_request.email,
        has_live=HAS_LIVE_TRADE_PLUGIN
    )


@fastapi_app.post("/get-config")
def get_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import get_config as gc

    return JSONResponse({
        'data': gc(json_request.current_config, has_live=HAS_LIVE_TRADE_PLUGIN)
    }, status_code=200)


@fastapi_app.post("/update-config")
def update_config(json_request: ConfigRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes.data_provider import update_config as uc

    uc(json_request.current_config)

    return JSONResponse({'message': 'Updated configurations successfully'}, status_code=200)


@fastapi_app.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket, token: str = Query(...)):
    from jesse.services.multiprocessing import process_manager

    if not authenticator.is_valid_token(token):
        return

    await websocket.accept()

    queue = Queue()
    ch, = await async_redis.psubscribe('channel:*')

    async def echo(q):
        while True:
            msg = await q.get()
            msg = json.loads(msg)
            msg['id'] = process_manager.get_client_id(msg['id'])
            await websocket.send_json(
                msg
            )

    async def reader(channel, q):
        async for ch, message in channel.iter():
            # modify id and set the one that the font-end knows
            await q.put(message)

    asyncio.get_running_loop().create_task(reader(ch, queue))
    asyncio.get_running_loop().create_task(echo(queue))

    try:
        while True:
            # just so WebSocketDisconnect would be raised on connection close
            await websocket.receive_text()
    except WebSocketDisconnect:
        await async_redis.punsubscribe('channel:*')
        print('Websocket disconnected')


# create a Click group
@click.group()
@click.version_option(pkg_resources.get_distribution("jesse").version)
def cli() -> None:
    pass


@cli.command()
@click.option(
    '--strict/--no-strict', default=True,
    help='Default is the strict mode which will raise an exception if the values for license is not set.'
)
def install_live(strict: bool) -> None:
    from jesse.services.installer import install
    install(HAS_LIVE_TRADE_PLUGIN, strict)


@cli.command()
def run() -> None:
    validate_cwd()

    # run all the db migrations
    from jesse.services.migrator import run as run_migrations
    import peewee
    try:
        run_migrations()
    except peewee.OperationalError:
        sleep_seconds = 10
        print(f"Database wasn't ready. Sleep for {sleep_seconds} seconds and try again.")
        time.sleep(sleep_seconds)
        run_migrations()

    # read port from .env file, if not found, use default
    from jesse.services.env import ENV_VALUES
    if 'APP_PORT' in ENV_VALUES:
        port = int(ENV_VALUES['APP_PORT'])
    else:
        port = 9000

    # run the main application
    uvicorn.run(fastapi_app, host="0.0.0.0", port=port, log_level="info")


@fastapi_app.post('/general-info')
def general_info(authorization: Optional[str] = Header(None)) -> JSONResponse:
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes import data_provider

    try:
        data = data_provider.get_general_info(has_live=HAS_LIVE_TRADE_PLUGIN)
    except Exception as e:
        return JSONResponse({
            'error': str(e)
        }, status_code=500)

    return JSONResponse(
        data,
        status_code=200
    )


@fastapi_app.post('/import-candles')
def import_candles(request_json: ImportCandlesRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
    from jesse.services.multiprocessing import process_manager

    validate_cwd()

    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.modes import import_candles_mode

    process_manager.add_task(
        import_candles_mode.run, 'candles-' + str(request_json.id), request_json.exchange, request_json.symbol,
        request_json.start_date, True
    )

    return JSONResponse({'message': 'Started importing candles...'}, status_code=202)


@fastapi_app.delete("/import-candles")
def cancel_import_candles(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    from jesse.services.multiprocessing import process_manager

    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    process_manager.cancel_process('candles-' + request_json.id)

    return JSONResponse({'message': f'Candles process with ID of {request_json.id} was requested for termination'}, status_code=202)


@fastapi_app.post("/backtest")
def backtest(request_json: BacktestRequestJson, authorization: Optional[str] = Header(None)):
    from jesse.services.multiprocessing import process_manager

    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    validate_cwd()

    from jesse.modes.backtest_mode import run as run_backtest

    process_manager.add_task(
        run_backtest,
        'backtest-' + str(request_json.id),
        request_json.debug_mode,
        request_json.config,
        request_json.routes,
        request_json.extra_routes,
        request_json.start_date,
        request_json.finish_date,
        None,
        request_json.export_chart,
        request_json.export_tradingview,
        request_json.export_full_reports,
        request_json.export_csv,
        request_json.export_json
    )

    return JSONResponse({'message': 'Started backtesting...'}, status_code=202)


@fastapi_app.post("/optimization")
async def optimization(request_json: OptimizationRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.multiprocessing import process_manager

    validate_cwd()

    from jesse.modes.optimize_mode import run as run_optimization

    process_manager.add_task(
        run_optimization,
        'optimize-' + str(request_json.id),
        request_json.debug_mode,
        request_json.config,
        request_json.routes,
        request_json.extra_routes,
        request_json.start_date,
        request_json.finish_date,
        request_json.optimal_total,
        request_json.export_csv,
        request_json.export_json
    )

    # optimize_mode(start_date, finish_date, optimal_total, cpu, csv, json)

    return JSONResponse({'message': 'Started optimization...'}, status_code=202)


@fastapi_app.delete("/optimization")
def cancel_optimization(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.multiprocessing import process_manager

    process_manager.cancel_process('optimize-' + request_json.id)

    return JSONResponse({'message': f'Optimization process with ID of {request_json.id} was requested for termination'}, status_code=202)


@fastapi_app.get("/download/{mode}/{file_type}/{session_id}")
def download(mode: str, file_type: str, session_id: str, token: str = Query(...)):
    if not authenticator.is_valid_token(token):
        return authenticator.unauthorized_response()

    from jesse.modes import data_provider

    return data_provider.download_file(mode, file_type, session_id)


@fastapi_app.delete("/backtest")
def cancel_backtest(request_json: CancelRequestJson, authorization: Optional[str] = Header(None)):
    if not authenticator.is_valid_token(authorization):
        return authenticator.unauthorized_response()

    from jesse.services.multiprocessing import process_manager

    process_manager.cancel_process('backtest-' + request_json.id)

    return JSONResponse({'message': f'Backtest process with ID of {request_json.id} was requested for termination'}, status_code=202)


@fastapi_app.on_event("shutdown")
def shutdown_event():
    from jesse.services.db import database
    database.close_connection()


if HAS_LIVE_TRADE_PLUGIN:
    from jesse.services.web import fastapi_app, LiveRequestJson, LiveCancelRequestJson, GetCandlesRequestJson, \
        GetLogsRequestJson, GetOrdersRequestJson
    from jesse.services import auth as authenticator

    @fastapi_app.post("/live")
    def live(request_json: LiveRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
        if not authenticator.is_valid_token(authorization):
            return authenticator.unauthorized_response()

        from jesse import validate_cwd

        # dev_mode is used only by developers so it doesn't have to be a supported parameter
        dev_mode: bool = False

        validate_cwd()

        # execute live session
        from jesse_live import live_mode
        from jesse.services.multiprocessing import process_manager

        trading_mode = 'livetrade' if request_json.paper_mode is False else 'papertrade'

        process_manager.add_task(
            live_mode.run,
            f'{trading_mode}-' + str(request_json.id),
            request_json.debug_mode,
            dev_mode,
            request_json.config,
            request_json.routes,
            request_json.extra_routes,
            trading_mode,
        )

        mode = 'live' if request_json.paper_mode is False else 'paper'

        return JSONResponse({'message': f"Started {mode} trading..."}, status_code=202)


    @fastapi_app.delete("/live")
    def cancel_backtest(request_json: LiveCancelRequestJson, authorization: Optional[str] = Header(None)):
        if not authenticator.is_valid_token(authorization):
            return authenticator.unauthorized_response()

        from jesse.services.multiprocessing import process_manager

        trading_mode = 'livetrade' if request_json.paper_mode is False else 'papertrade'

        process_manager.cancel_process(f'{trading_mode}-' + request_json.id)

        return JSONResponse({'message': f'Live process with ID of {request_json.id} terminated.'}, status_code=200)


    @fastapi_app.post('/get-candles')
    def get_candles(json_request: GetCandlesRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
        if not authenticator.is_valid_token(authorization):
            return authenticator.unauthorized_response()

        from jesse import validate_cwd

        validate_cwd()

        from jesse.modes.data_provider import get_candles as gc

        arr = gc(json_request.exchange, json_request.symbol, json_request.timeframe)

        return JSONResponse({
            'id': json_request.id,
            'data': arr
        }, status_code=200)


    @fastapi_app.post('/get-logs')
    def get_logs(json_request: GetLogsRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
        if not authenticator.is_valid_token(authorization):
            return authenticator.unauthorized_response()

        from jesse_live.services.data_provider import get_logs as gl

        arr = gl(json_request.session_id, json_request.type)

        return JSONResponse({
            'id': json_request.id,
            'data': arr
        }, status_code=200)


    @fastapi_app.post('/get-orders')
    def get_orders(json_request: GetOrdersRequestJson, authorization: Optional[str] = Header(None)) -> JSONResponse:
        if not authenticator.is_valid_token(authorization):
            return authenticator.unauthorized_response()

        from jesse_live.services.data_provider import get_orders as go

        arr = go(json_request.session_id)

        return JSONResponse({
            'id': json_request.id,
            'data': arr
        }, status_code=200)


# Mount static files.Must be loaded at the end to prevent overlapping with API endpoints
fastapi_app.mount("/", StaticFiles(directory=f"{JESSE_DIR}/static"), name="static")



import numpy as np
import pandas as pd
import pytest

from jesse import utils
from tests.data.test_candles_indicators import test_candles_19


def test_anchor_timeframe():
    assert utils.anchor_timeframe('1m') == '5m'
    assert utils.anchor_timeframe('3m') == '15m'
    assert utils.anchor_timeframe('5m') == '30m'
    assert utils.anchor_timeframe('15m') == '2h'
    assert utils.anchor_timeframe('30m') == '3h'
    assert utils.anchor_timeframe('1h') == '4h'
    assert utils.anchor_timeframe('2h') == '6h'
    assert utils.anchor_timeframe('3h') == '1D'
    assert utils.anchor_timeframe('4h') == '1D'
    assert utils.anchor_timeframe('6h') == '1D'


def test_crossed():
    candles = np.array(test_candles_19)
    cross_100 = utils.crossed(candles[:, 2], 100)
    assert cross_100 == False
    cross_120 = utils.crossed(candles[:, 2], 120)
    assert cross_120 == True
    cross_120 = utils.crossed(candles[:, 2], 120, direction="below")
    assert cross_120 == True
    cross_120 = utils.crossed(candles[:, 2], 120, direction="above")
    assert cross_120 == False
    seq_cross_200 = utils.crossed(candles[:, 2], 200, direction="below", sequential=True)
    assert seq_cross_200[-5] == True
    seq_cross_200 = utils.crossed(candles[:, 2], 200, direction="above", sequential=True)
    assert seq_cross_200[-5] == False
    seq_cross_120 = utils.crossed(candles[:, 2], 120, sequential=True)
    assert seq_cross_120[-1] == True
    array_array_cross_above = utils.crossed(np.array([1., 2, 3, 4, 5, 6]), np.array([3., 3, 3, 3, 3, 3]),
                                            direction="above",
                                            sequential=True)
    assert array_array_cross_above[-3] == True
    array_array_cross_below = utils.crossed(np.array([1., 2, 3, 2, 1, 6]), np.array([3., 3, 3, 3, 3, 3]),
                                            direction="below",
                                            sequential=True)
    assert array_array_cross_below[-3] == True
    array_array_cross = utils.crossed(np.array([1., 2, 3, 4, 1, 2]), np.array([3., 3, 3, 3, 3, 3]),
                                      sequential=True)
    assert array_array_cross[-3] == True
    assert array_array_cross[-2] == True


def test_estimate_risk():
    assert utils.estimate_risk(100, 80) == 20


def test_limit_stop_loss():
    assert utils.limit_stop_loss(100, 105, 'short', 10) == 105
    assert utils.limit_stop_loss(100, 115, 'short', 10) == 110
    assert utils.limit_stop_loss(100, 95, 'long', 10) == 95
    assert utils.limit_stop_loss(100, 85, 'long', 10) == 90

    with pytest.raises(TypeError):
        utils.limit_stop_loss(100, 85, 'long', 'invalid_input')
        utils.limit_stop_loss('invalid_input', 105, 'short', 10)
        utils.limit_stop_loss(100, 'invalid_input', 'short', 10)
        utils.limit_stop_loss(100, 105, 123, 10)


def test_numpy_to_pandas():
    candles = np.array(test_candles_19)
    columns = ["Date", "Open", "Close", "High", "Low", "Volume"]
    df = pd.DataFrame(data=candles, index=pd.to_datetime(candles[:, 0], unit="ms"), columns=columns)
    df["Date"] = pd.to_datetime(df["Date"], unit="ms")

    ohlcv = utils.numpy_candles_to_dataframe(candles, name_date="Date", name_open="Open", name_high="High",
                                             name_low="Low", name_close="Close", name_volume="Volume")

    pd.testing.assert_frame_equal(df, ohlcv)


def test_qty_to_size():
    assert utils.qty_to_size(2, 50) == 100
    assert utils.qty_to_size(2, 49) == 98

    with pytest.raises(TypeError):
        utils.qty_to_size(-10, 'invalid_input')
        utils.qty_to_size('invalid_input', -10)
    with pytest.raises(TypeError):
        utils.qty_to_size(-10, None)
        utils.qty_to_size(None, -10)


def test_risk_to_qty():
    # long
    assert utils.risk_to_qty(10000, 1, 100, 80) == 5
    # short
    assert utils.risk_to_qty(10000, 1, 80, 100) == 5

    # should not return more than maximum capital. Expect 100 instead of 125
    assert utils.risk_to_qty(10000, 5, 100, 96) == 100

    # when fee is included
    assert utils.risk_to_qty(10000, 1, 100, 80, precision=3, fee_rate=0.001) == 4.97


def test_risk_to_size():
    assert round(utils.risk_to_size(10000, 1, 0.7, 8.6)) == 1229

    with pytest.raises(TypeError):
        utils.risk_to_size(10000, 1, 0.7, None)
        utils.risk_to_size(10000, 1, None, 8.6)
        utils.risk_to_size(10000, None, 0.7, 8.6)
        utils.risk_to_size(None, 1, 0.7, 8.6)

    # should not return more than maximum capital
    assert utils.risk_to_size(10000, 5, 4, 100) == 10000


def test_size_to_qty():
    assert utils.size_to_qty(100, 50) == 2
    assert utils.size_to_qty(100, 49, precision=3) == 2.04

    with pytest.raises(TypeError):
        utils.size_to_qty(100, 'invalid_input')
        utils.size_to_qty('invalid_input', 100)
    with pytest.raises(TypeError):
        utils.size_to_qty(100, None)
        utils.size_to_qty(None, 100)

    # when fee is included
    assert utils.size_to_qty(100, 50, fee_rate=0.001) == 1.994


def test_sum_floats():
    assert utils.sum_floats(9.71, 9.813) == 19.523
    assert utils.sum_floats(-1.123, -1.2) == -2.323
    assert utils.sum_floats(1.19, -1.2) == -0.01
    assert utils.sum_floats(-1.19, 1.2) == 0.01


def test_subtract_floats():
    assert utils.subtract_floats(9.813, 9.71) == 0.103
    assert utils.subtract_floats(-1.123, 1.2) == -2.323
    assert utils.subtract_floats(1.123, -1.2) == 2.323
    assert utils.subtract_floats(-1.123, -1.2) == 0.077


def test_prices_to_returns():
    series = np.array([50, 10, 100, 25])
    pct = utils.prices_to_returns(series)
    np.testing.assert_array_equal(pct, np.array([np.nan, -80., 900., -75.]))


def test_combinations_without_repeat():
    a = np.array([4, 2, 9, 1, 3])
    b = utils.combinations_without_repeat(a)
    np.testing.assert_array_equal(b, np.array([[4, 2],
                                               [4, 9],
                                               [4, 1],
                                               [4, 3],
                                               [2, 4],
                                               [2, 9],
                                               [2, 1],
                                               [2, 3],
                                               [9, 4],
                                               [9, 2],
                                               [9, 1],
                                               [9, 3],
                                               [1, 4],
                                               [1, 2],
                                               [1, 9],
                                               [1, 3],
                                               [3, 4],
                                               [3, 2],
                                               [3, 9],
                                               [3, 1]]))


def test_wavelet_denoising():
    candles = np.array(test_candles_19)
    denoised = utils.wavelet_denoising(candles[:, 2], wavelet="sym4", level=1, mode='symmetric', smoothing_factor=2)
    assert len(candles) == len(denoised)



from collections import namedtuple

import numpy as np
import talib as ta

from jesse.helpers import get_candle_source, slice_candles

Wavetrend = namedtuple('Wavetrend', ['wt1', 'wt2', 'wtCrossUp', 'wtCrossDown', 'wtOversold', 'wtOverbought', 'wtVwap'])


# Wavetrend indicator ported from:  https://www.tradingview.com/script/Msm4SjwI-VuManChu-Cipher-B-Divergences/
#                                   https://www.tradingview.com/script/2KE8wTuF-Indicator-WaveTrend-Oscillator-WT/
#
# buySignal = wtCross and wtCrossUp and wtOversold
# sellSignal = wtCross and wtCrossDown and wtOverbought
#
# See https://github.com/ysdede/lazarus3/blob/partialexit/strategies/lazarus3/__init__.py for working jesse.ai example.


def wt(candles: np.ndarray,
       wtchannellen: int = 9,
       wtaveragelen: int = 12,
       wtmalen: int = 3,
       oblevel: int = 53,
       oslevel: int = -53,
       source_type: str = "hlc3",
       sequential: bool = False) -> Wavetrend:
    """
    Wavetrend indicator

    :param candles: np.ndarray
    :param wtchannellen:  int - default: 9
    :param wtaveragelen: int - default: 12
    :param wtmalen: int - default: 3
    :param oblevel: int - default: 53
    :param oslevel: int - default: -53
    :param source_type: str - default: "hlc3"
    :param sequential: bool - default: False

    :return: Wavetrend
    """
    candles = slice_candles(candles, sequential)

    src = get_candle_source(candles, source_type=source_type)

    # wt
    esa = ta.EMA(src, wtchannellen)
    de = ta.EMA(abs(src - esa), wtchannellen)
    ci = (src - esa) / (0.015 * de)
    wt1 = ta.EMA(ci, wtaveragelen)
    wt2 = ta.SMA(wt1, wtmalen)

    wtVwap = wt1 - wt2
    wtOversold = wt2 <= oslevel
    wtOverbought = wt2 >= oblevel
    wtCrossUp = wt2 - wt1 <= 0
    wtCrossDown = wt2 - wt1 >= 0

    if sequential:
        return Wavetrend(wt1, wt2, wtCrossUp, wtCrossDown, wtOversold, wtOverbought, wtVwap)
    else:
        return Wavetrend(wt1[-1], wt2[-1], wtCrossUp[-1], wtCrossDown[-1], wtOversold[-1], wtOverbought[-1], wtVwap[-1])
